# -*- coding: utf-8 -*-
"""
Created on  by Ismail Baris
"""
from __future__ import division

from rspy.units.si_units import __values__, __unit__


class TestValuesUnit:
    def test_len(self):
        assert len(__values__) == len(__unit__)



# -*- coding: utf-8 -*-
"""
Created on 06.04.19 by ibaris
"""
from __future__ import division

import numpy as np

from rspy.angles import Angles

__all__ = ['Phase']


class Phase(Angles):
    """
    This Phase class defines the different phase functions to calculate the Radiative Transfer Based
    kernels. It also can integrate these phase functions.

    """

    def __init__(self, iza, vza, raa=None, iaa=None, vaa=None, m=2, br_ratio=1, hb_ratio=2, method='Ross',
                 normalize=False, nbar=0.0, angle_unit='DEG', orientation="uniform"):
        """
        This Phase class defines the different phase functions to calculate the Radiative Transfer Based
        kernels.

        Parameters
        ----------
        iza, vza, raa : int, float or ndarray
            Incidence (iza) and scattering (vza) zenith angle, as well as relative azimuth (raa) angle.
        normalize : boolean, optional
            Set to 'True' to make kernels 0 at nadir view illumination. Since all implemented kernels are normalized
            the default value is False.
        nbar : float, optional
            The sun or incidence zenith angle at which the isotropic term is set
            to if normalize is True. The default value is 0.0.
        angle_unit : {'DEG', 'RAD'}, optional
            * 'DEG': All input angles (iza, vza, raa) are in [DEG] (default).
            * 'RAD': All input angles (iza, vza, raa) are in [RAD].

        Returns
        -------
        All returns are attributes! For more Attributes see also Kernel

        See Also
        --------
        Kernel

        Note
        ----
        Hot spot direction is vza == iza and raa = 0.0

        """

        super(Phase, self).__init__(iza=iza, vza=vza, raa=raa, iaa=iaa, vaa=vaa, normalize=normalize, nbar=nbar,
                                    angle_unit=angle_unit,
                                    align=True)

        self.__m, self.__br, self.__hb = self.align_with((m, br_ratio, hb_ratio))
        self.method = method

        self.__piza, self.__pvza = self.__calculate_projected_angles()
        self.__D = self.__calculate_distance()
        self.__O = self.__calculate_overlap()
        self.__phaang = self.compute_phase_angle(self.iza, self.vza, self.raa, self.method)

        if orientation == 'uniform' or orientation == 'vertical' or orientation == 'horizontal':
            self.orientation = orientation
            self.__phase_functions = {'uniform': self.__uniform,
                                      'vertical': self.__vertical,
                                      'horizontal': self.__horizontal}

        else:
            raise ValueError("'phase_function' must be 'uniform', 'horizontal' or 'vertical'. "
                             "The actual 'phase_function' is: {0}".format(orientation))

        if method == 'Ross' or method == 'Omari':
            self.method = method

        else:
            raise ValueError("method must be 'Ross' or 'Omari', method is:{0]".format(str(method)))

    # --------------------------------------------------------------------------------------------------------
    # Properties
    # --------------------------------------------------------------------------------------------------------
    @property
    def D(self):
        return self.__D

    @property
    def O(self):
        return self.__O

    @property
    def m(self):
        return self.__m

    @property
    def br(self):
        return self.__br

    @property
    def hb(self):
        return self.__hb

    @property
    def piza(self):
        return self.__piza

    @property
    def pvza(self):
        return self.__pvza

    @property
    def phaang(self):
        return self.__phaang

    # --------------------------------------------------------------------------------------------------------
    # Callable Methods
    # --------------------------------------------------------------------------------------------------------
    @staticmethod
    def compute_phase_angle(iza, vza, phi, method='Ross'):
        """
        A method to calculate the phase angle components between the angle.

        Parameters
        ----------
        iza, vza, phi : int, float or ndarray
            Incidence (iza) and scattering (vza) zenith angle, as well as relative azimuth angle in a range between
            0 and 2pi.
        method : {'Ross', 'Omari'} (default = 'Ross'), optional
            * 'Ross': Phase angle calculation based on :cite:`Ross.1981` (default).
            * 'Omari': Phase angle calculation based on :cite:`Omari.2009`.

        Returns
        -------
        phaang: ndarray
            Phase angle in [RAD].
        cosphaang: ndarray
            Cosine of phase angle in [RAD].
        sinphaang: ndarray
            Sine of phase angle in [Rad].

        """

        if method == 'Ross':
            cosphaang = np.cos(iza) * np.cos(vza) + np.sin(iza) * np.sin(vza) * np.cos(phi)

            # better check the bounds...  just to be safe
            w = np.where(cosphaang < -1)[0]
            cosphaang[w] = -1.0
            w = np.where(cosphaang > 1)[0]
            cosphaang[w] = 1.0
            phaang = np.arccos(cosphaang)

        elif method == 'Omari':
            phaang = np.arccos(np.cos(iza) * np.cos(vza))

        else:
            raise ValueError("method must be 'Ross' or 'Omari', method is:{0]".format(str(method)))

        return phaang

    def compute_Z(self, iza=None, vza=None, iaa=None, vaa=None, method='Ross'):
        """
        Compute the phase function for current setup.

        Parameters
        ----------
        iza, vza, raa : int, float or ndarray
            Incidence (iza) and scattering (vza) zenith angle, as well as relative azimuth (raa) angle.

        Returns
        -------
        out : ndarray
            Phase values.
        """
        if iaa is None and vaa is None:
            phi = self.phi
        else:
            phi = iaa - vaa

        if method == 'Ross':
            return self.__phase_functions[self.orientation](iza, vza, phi)

        elif method == 'Li':
            return self.__li_sparse(iza, vza, phi)

        else:
            raise ValueError("'method' must be 'Ross' or 'Li'. "
                             "The actual value is: {0}".format(self.orientation))

    # --------------------------------------------------------------------------------------------------------
    # Private Methods
    # --------------------------------------------------------------------------------------------------------
    def __calculate_projected_angles(self):
        ti = self.__br * np.tan(self.iza)
        tv = self.__br * np.tan(self.vza)

        wi = np.where(ti < 0.)[0]
        wv = np.where(tv < 0.)[0]
        ti[wi] = 0.0
        tv[wv] = 0.0

        piza = np.arctan(ti)
        pvza = np.arctan(tv)

        return piza, pvza

    def __calculate_distance(self):
        """
        A method to calculate the distance component of geometric kernels.

        Parameters
        ----------
        iza, vza, raa : int, float or ndarray
            Incidence (iza) and scattering (vza) zenith angle, as well as relative azimuth (raa) angle.

        Returns
        --------
        out: ndarray
            Square root of inner_distance.
        """
        inner_distance = np.power(np.tan(self.__piza), 2) + np.power(np.tan(self.__pvza), 2) - 2. * np.tan(
            self.__piza) * np.tan(self.__pvza) * np.cos(self.phi)

        w = np.where(inner_distance < 0)[0]
        inner_distance[w] = 0.0

        return np.sqrt(inner_distance)

    def __calculate_overlap(self):
        cost = self.__hb * np.sqrt(
            self.__D ** 2 + np.tan(self.__pvza) ** 2 * np.tan(self.__piza) ** 2 * np.sin(self.phi) ** 2) / self.B

        w = np.where(cost < -1)[0]
        cost[w] = -1.0
        w = np.where(cost > 1.0)[0]
        cost[w] = 1.0
        tvar = np.arccos(cost)
        sint = np.sin(tvar)
        overlap = (1 / np.pi) * (tvar - sint * cost) * self.B
        w = np.where(overlap < 0)[0]
        overlap[w] = 0.0

        tvar = np.arccos(cost)
        sint = np.sin(tvar)
        overlap = (1 / np.pi) * (tvar - sint * cost) * self.B
        w = np.where(overlap < 0)[0]
        overlap[w] = 0.0

        return overlap

    def __uniform(self, iza=None, vza=None, phi=None):
        iza = self.iza if iza is None else iza
        vza = self.vza if vza is None else vza
        phi = self.phi if phi is None else phi

        phaang = self.compute_phase_angle(iza, vza, phi, self.method)
        m = 2 if self.__m is None else self.__m

        return ((np.pi / m) - phaang) * np.cos(phaang) + np.sin(phaang)

    def __li_sparse(self, piza=None, pvza=None, phi=None):
        piza = self.piza if piza is None else piza
        pvza = self.pvza if pvza is None else pvza
        phi = self.phi if phi is None else phi

        phaang = self.compute_phase_angle(piza, pvza, phi, self.method)

        return 1 / 2 * (1 + np.cos(phaang))

    def __vertical(self, iza=None, vza=None, phi=None):
        raise NotImplementedError("Only uniform phase function is implemented yet.")

    def __horizontal(self, iza=None, vza=None, phi=None):
        raise NotImplementedError("Only uniform phase function is implemented yet.")



# -*- coding: utf-8 -*-
"""
Created on  by Ismail Baris
"""
from __future__ import division

from rspy.__version__ import __version__, version_info


class TestVersion:
    def test_version(self):
        assert __version__ == '.'.join(map(str, version_info))



# -*- coding: utf-8 -*-
"""
Created on 06.04.19 by ibaris
"""
from __future__ import division

import numpy as np
import pytest

from rspy import Phase, same_len

n = 100

iza, vza, iaa, vaa, nbar = np.random.uniform(1, np.pi, n), np.random.uniform(1, np.pi, n), \
                           np.random.uniform(1, 2 * np.pi, n), np.random.uniform(1, 2 * np.pi, n), \
                           np.random.uniform(1, np.pi / 4, 1)

izaDeg, vzaDeg, iaaDeg, vaaDeg, nbarDeg = np.rad2deg(iza), np.rad2deg(vza), np.rad2deg(iaa), np.rad2deg(
    vaa), np.rad2deg(nbar)

raa = iaa - vaa
raaDeg = iaaDeg - vaaDeg

degitems = [izaDeg, vzaDeg, iaaDeg, vaaDeg, raaDeg]
raditems = [iza, vza, iaa, vaa, raa]

degitems_key = ["izaDeg", "vzaDeg", "iaaDeg", "vaaDeg", "raaDeg"]
raditems_key = ["iza", "vza", "iaa", "vaa", "raa"]


class TestExceptions:
    def test_raa_iaa_vaa_convention(self):
        with pytest.raises(ValueError):
            ang = Phase(iza=iza, vza=vza, iaa=iaa)

        with pytest.raises(ValueError):
            ang = Phase(iza=iza, vza=vza, vaa=vaa)

        with pytest.raises(ValueError):
            ang = Phase(iza=iza, vza=vza)

        with pytest.raises(AssertionError):
            ang = Phase(iza=iza, vza=vza, raa=raa, iaa=iaa, vaa=vaa)

    def test_valid_units(self):
        with pytest.raises(ValueError):
            ang = Phase(iza=iza, vza=vza, iaa=iaa, vaa=vaa, angle_unit='degg')


class TestInit:
    def test_alignment(self):
        ang = Phase(iza=10, vza=vzaDeg, iaa=iaaDeg, vaa=vaaDeg)

        assert same_len((ang.iza, ang.vza, ang.iaa, ang.vaa, ang.raa, ang.alpha, ang.beta))
        assert same_len((ang.izaDeg, ang.vzaDeg, ang.iaaDeg, ang.vaaDeg, ang.raaDeg, ang.alphaDeg,
                         ang.betaDeg))

    def test_raa_flags(self):
        ang = Phase(iza=izaDeg, vza=vzaDeg, iaa=iaaDeg, vaa=vaaDeg)
        assert np.allclose(ang.raa, ang.iaa - ang.vaa)

        ang = Phase(iza=izaDeg, vza=vzaDeg, raa=iaaDeg)
        assert np.allclose(ang.iaa, 0)
        assert np.allclose(ang.vaa, 0)

    def test_negative_angles(self):
        ang = Phase(iza=-izaDeg, vza=-vzaDeg, iaa=iaaDeg, vaa=vaaDeg)

        for i, item in enumerate(degitems):
            if 'aa' in degitems_key[i]:
                itemaa = item + 180

                assert np.allclose(ang[degitems_key[i]], itemaa)
            else:
                assert np.allclose(ang[degitems_key[i]], item)

        for i, item in enumerate(raditems):
            if 'aa' in raditems_key[i]:
                itemaa = item + np.pi

                assert np.allclose(ang[raditems_key[i]], itemaa)
            else:
                assert np.allclose(ang[raditems_key[i]], item)



# -*- coding: utf-8 -*-
"""
Created on  by Ismail Baris
"""
from __future__ import division

import operator

import numpy as np
from rspy import Units

n = 30

__OPERAND__ = ['*', '/', '+', '-', '**']
__OPERAND_OBJECT__ = [operator.mul, operator.truediv, operator.add, operator.sub, operator.pow]

__OPERATOR__ = dict(zip(__OPERAND__, __OPERAND_OBJECT__))


class TestDimensionAndUnitMethods:
    def test_unit_isnone(self):
        pass

    def test_dim_isnone(self):
        unit_str = Units.length.keys()

        for item in unit_str:
            unit = Units.units[item] ** 4
            dim = Units.dim_isnone(unit)

            assert dim

    def test_dim_isone(self):
        unit_str = Units.other.keys()

        for item in unit_str:
            unit = Units.units[item]
            dim = Units.dim_isone(unit)

            assert dim


class TestExpressionMethods:
    def test_isexpr_scalar(self):
        for x in range(n):
            operand = np.random.random()
            operator = np.random.choice(__OPERAND_OBJECT__)
            ind = np.random.choice(len(Units.units.values()))
            values = list(Units.units.values())
            unit = values[ind]

            expr = operator(operand, unit)

            assert Units.isexpr(expr)

    def test_isexpr_array(self):
        for x in range(n):
            i, j, k = np.random.randint(1, 10), np.random.randint(1, 10), np.random.randint(1, 10)
            operand = np.random.random((i, j, k))
            operator = np.random.choice(__OPERAND_OBJECT__)
            ind = np.random.choice(len(Units.units.values()))
            values = list(Units.units.values())
            unit = values[ind]

            expr = operator(operand, unit)

            assert Units.isexpr(expr)



# -*- coding: utf-8 -*-
"""
Created on 06.04.19 by ibaris
"""
from __future__ import division

from numpy import errstate, nan_to_num, log10

__all__ = ['dB', 'linear']


def dB(x):
    """
    Convert a linear value to dB.
    """
    with errstate(invalid='ignore'):
        return nan_to_num(10 * log10(x))


def linear(x):
    """
    Convert a dB value in linear.
    """
    return 10 ** (x / 10)



# -*- coding: utf-8 -*-
"""
Created on 01.04.2019 by Ismail Baris
"""
from __future__ import division

from numpy import ndarray

from rspy.units.auxiliary import (__NONE_TYPE_UNITS__, __NONE_TYPE_DIMS__, __SYMPY_CLASSES__, One, Zero)

__all__ = ['unit_isnone', 'dim_isnone', 'dim_isone', 'dim_iszero', 'isexpr']


def unit_isnone(unit):
    """
    Check if a unit has a None-Typed object.

    Parameters
    ----------
    unit : object
        Unit expression.

    Returns
    -------
    bool
    """
    if unit in __NONE_TYPE_UNITS__:
        return True

    return False


def dim_isnone(unit):
    """
    Check if a dimension of a unit has a None-Typed object.

    Parameters
    ----------
    unit : object
        Unit expression.

    Returns
    -------
    bool
    """
    try:
        if unit.dimension.name in __NONE_TYPE_DIMS__:
            return True
        return False

    except AttributeError:
        return True


def dim_isone(unit):
    """
    Check if a dimension of a unit is One.

    Parameters
    ----------
    unit : object
        Unit expression.

    Returns
    -------
    bool
    """
    try:
        if unit.dimension.name == One:
            return True

        return False

    except AttributeError:
        return False


def dim_iszero(unit):
    """
    Check if a dimension of a unit is Zero.

    Parameters
    ----------
    unit : object
        Unit expression.

    Returns
    -------
    bool
    """
    try:
        if unit.dimension.name == Zero:
            return True

        return False

    except AttributeError:
        return False


def isexpr(value):
    """
    Check if a object is sympy expression.

    Parameters
    ----------
    value : object
        Expression.

    Returns
    -------
    bool
    """
    if isinstance(value, __SYMPY_CLASSES__):
        return True

    elif isinstance(value, ndarray):
        if isinstance(value.flatten()[0], __SYMPY_CLASSES__):
            return True

    return False



# -*- coding: utf-8 -*-
"""
Created on 07.04.19 by ibaris
"""
from __future__ import division
from rspy.intensity import *
import numpy as np

n = 50


class TestBRFBRDF:
    def test_db_lin(self):
        for x in range(n):
            BRDF = np.random.uniform(0.00001, 0.1, n)
            dB_val = dB(BRDF)
            lin_val = linear(dB_val)

            assert np.allclose(BRDF, lin_val)



# -*- coding: utf-8 -*-
"""
Created on 06.04.19 by ibaris
"""
from __future__ import division

import numpy as np

import rspy.constants as const
from rspy.ancillary import d2r
from rspy.auxiliary import check_angle_unit
from rspy.auxiliary.types import __UNIT_RAD__, __UNIT_DEG__
from rspy.intensity.utility import dB, linear

__all__ = ['Intensity']


class Intensity(object):
    """

    Attributes
    ----------
    I : array_like or respy.unit.quantity.Quantity
        Intensity (BRDF) value.
    BRF : array_like or respy.unit.quantity.Quantity
        BRF value.
    BSC : array_like or respy.unit.quantity.Quantity
        BSC value.
    BSCdB : array_like or respy.unit.quantity.Quantity
        BSC value in dB.
    value : array_like
        All values in an array of type np.array([I, BRF, BSC, BSCdB])

    Methods
    -------
    Conversion.dB : Convert linear to dB
    Conversion.linear : Convert dB to linear.

    """

    def __init__(self, value, vza=None, value_unit="BRDF", angle_unit='RAD'):
        """
        Conversion of BRDF, BRF, BSC and BSC in dB.

        Parameters
        ----------
        value : float, array_like
            Input value in BRDF, BRF, BSC or BSCdB. See parameter value_unit.
        vza : int, float, array_like
            Viewing zenith angle in DEG or RAD. See parameter angle_unit.
        value_unit : {'I', 'BRDF', 'BRF', 'BSC', 'BSCdB', 'BSCdb', 'brdf', 'brf', 'bsc', 'bscdb'}
            The unit of input value:
            * I or BRDF : Bidirectional Reflectance Distribution Function (Intensity) (default).
            * BRF : Bidirectional Reflectance Factor.
            * BSC : Back Scattering Coefficient (no unit).
            * BSCdB : Back Scattering Coefficient in dB.
        angle_unit : {'DEG', 'RAD', 'deg', 'rad'}, optional
            * 'DEG': Input angle in [DEG].
            * 'RAD': Input angle  in [RAD] (default).

        """
        self.value_unit = value_unit
        self.angle_unit = angle_unit

        if check_angle_unit(angle_unit):
            pass
        else:
            raise ValueError("angle_unit must be {0} or {1}, "
                             "but the actual angle_unit is: {2}".format(str(__UNIT_RAD__),
                                                                        str(__UNIT_DEG__),
                                                                        str(angle_unit)))

        value = np.asarray(value)
        value = np.atleast_1d(value)

        if self.value_unit == "BRDF":
            self.I = value
            self.BRF = Intensity.BRDF_to_BRF(value)

            if vza is not None:
                self.BSC = Intensity.BRDF_to_BSC(self.I, vza, self.angle_unit)
                self.BSCdB = dB(Intensity.BRDF_to_BSC(self.I, vza, self.angle_unit))

            else:
                self.BSC = np.zeros_like(value)
                self.BSCdB = np.zeros_like(value)

        elif self.value_unit == "BSC":
            self.BSC = value
            self.BSCdB = dB(value)

            if vza is not None:
                self.I = Intensity.BSC_to_BRDF(self.BSC, vza, self.angle_unit)
                self.BRF = Intensity.BRDF_to_BRF(self.I)
            else:
                self.I = np.zeros_like(value)
                self.BRF = np.zeros_like(value)

        elif self.value_unit == "BSCdB":
            self.BSCdB = value
            self.BSC = linear(value)

            if vza is not None:
                self.I = Intensity.BSC_to_BRDF(self.BSC, vza, self.angle_unit)
                self.BRF = Intensity.BRDF_to_BRF(self.I)
            else:
                self.I = np.zeros_like(value)
                self.BRF = np.zeros_like(value)

        elif self.value_unit == "BRF":
            self.BRF = value
            self.I = value / const.pi

            if vza is not None:
                self.BSC = Intensity.BRDF_to_BSC(self.I, vza, self.angle_unit)
                self.BSCdB = dB(Intensity.BRDF_to_BSC(self.I, vza, self.angle_unit))

            else:
                self.BSC = np.zeros_like(value)
                self.BSCdB = np.zeros_like(value)

        else:
            raise ValueError("the unit of value must be 'BRDF', 'BRF', 'BSC' or 'BSCdB'")

        self.array = np.array([self.I, self.BRF, self.BSC, self.BSCdB])
        self.dict = {'I': self.I,
                     'BRF': self.BRF,
                     'BSC': self.BSC,
                     'BSCdB': self.BSCdB}

    # --------------------------------------------------------------------------------------------------------
    # Magic Methods
    # --------------------------------------------------------------------------------------------------------
    def __repr__(self):
        prefix = '<{0} '.format(self.__class__.__name__)
        sep = ','

        arrstr_bsc = np.array2string(self.BSC,
                                     separator=sep,
                                     prefix=prefix)

        arrstr_dB = np.array2string(self.BSCdB,
                                    separator=sep,
                                    prefix=prefix)

        arrstr_I = np.array2string(self.I,
                                   separator=sep,
                                   prefix=prefix)

        arrstr_BRF = np.array2string(self.BRF,
                                     separator=sep,
                                     prefix=prefix)

        bsc = '{0}{1} Backscattering Coefficient in [{2}]>'.format(prefix, arrstr_bsc, '[-]')
        bscdb = '{0}{1} Backscattering Coefficient in [{2}]>'.format(prefix, arrstr_dB, '[dB]')
        I = '{0}{1} Intensity in [{2}]>'.format(prefix, arrstr_I, '[-]')
        BRF = '{0}{1} Bidirectional Reflectance Factor in [{2}]>'.format(prefix, arrstr_BRF, '[-]')

        return I + '\n' + BRF + '\n' + bsc + '\n' + bscdb

    def __getitem__(self, item):
        if isinstance(item, str):
            try:
                return self.dict[item]
            except KeyError:
                raise KeyError("{} is not a valid value. Use `keys()` method to see all available values".format(item))

        return self.array[item]

    def __dir__(self):
        return self.keys()

    # --------------------------------------------------------------------------------------------------------
    # Callable Methods
    # --------------------------------------------------------------------------------------------------------
    @staticmethod
    def BRDF_to_BRF(BRDF):
        """
        Convert a BRDF into a BRF.

        Parameters
        ----------
        BRDF : int, float or array_like
            BRDF value.

        Returns
        -------
        BRF value : int, float or array_like

        """
        return const.pi * BRDF

    @staticmethod
    def BSC_to_BRDF(BSC, vza, angle_unit='RAD'):
        """
        Convert a Radar Backscatter Coefficient (BSC) into a BRDF.

        Parameters
        ----------
        BSC : int, float or array_like
            Radar Backscatter Coefficient (sigma 0).
        vza : int, float or array_like
            View or scattering zenith angle.
        angle_unit : {'DEG', 'RAD'} (default = 'RAD'), optional
            * 'DEG': All input angles (iza, vza, raa) are in [DEG].
            * 'RAD': All input angles (iza, vza, raa) are in [RAD].

        Returns
        -------
        BRDF value : int, float or array_like

        """
        if check_angle_unit(angle_unit):
            pass
        else:
            raise ValueError("angle_unit must be {0} or {1}, "
                             "but the actual angle_unit is: {2}".format(str(__UNIT_RAD__),
                                                                        str(__UNIT_DEG__),
                                                                        str(angle_unit)))

        vza = np.asarray(vza)
        vza = np.atleast_1d(vza)

        if angle_unit in __UNIT_RAD__:
            mu = np.cos(vza)
        else:
            mu = np.cos(d2r(vza))

        if BSC.ndim != vza.ndim:
            if BSC.shape[0] == vza.shape[0]:
                BRDF = np.empty_like(BSC)

                for i in range(BSC.shape[0]):
                    BRDF[i] = BSC[i] / (mu[i] * (4 * const.pi))

                return BRDF
            else:
                raise AssertionError("Inputs must have the same length.")

        return BSC / (mu * (4 * const.pi))

    @staticmethod
    def BRDF_to_BSC(BRDF, vza, angle_unit='RAD'):
        """
        Convert a BRDF in to a Radar Backscatter Coefficient (BSC).

        Note
        -----
        Hot spot direction is vza == iza and raa = 0.0

        Parameters
        ----------
        BRDF : int, float or array_like
            Intensity as a BRDF.
        vza : int, float or array_like
            View or scattering zenith angle.
        angle_unit : {'DEG', 'RAD'} (default = 'RAD'), optional
            * 'DEG': All input angles (iza, vza, raa) are in [DEG].
            * 'RAD': All input angles (iza, vza, raa) are in [RAD].

        Returns
        -------
        BRDF value : int, float or array_like

        """
        if check_angle_unit(angle_unit):
            pass
        else:
            raise ValueError("angle_unit must be {0} or {1}, "
                             "but the actual angle_unit is: {2}".format(str(__UNIT_RAD__),
                                                                        str(__UNIT_DEG__),
                                                                        str(angle_unit)))
        vza = np.asarray(vza)
        vza = np.atleast_1d(vza)

        if angle_unit in __UNIT_RAD__:
            mu = np.cos(vza)
        else:
            mu = np.cos(d2r(vza))

        if BRDF.ndim != vza.ndim:
            if BRDF.shape[0] == vza.shape[0]:
                BSC = np.empty_like(BRDF)

                for i in range(BSC.shape[0]):
                    BSC[i] = BRDF[i] * mu[i] * 4 * const.pi

                return BSC
            else:
                raise AssertionError("Inputs must have the same length.")

        return BRDF * mu * 4 * const.pi

    def keys(self):
        return self.dict.keys()

    def values(self):
        return self.dict.values()



# -*- coding: utf-8 -*-
"""
Created on 01.04.2019 by Ismail Baris
"""
from __future__ import division

import sympy.physics.units as sympy_units
from sympy import S
from sympy.physics.units.quantities import Quantity as sQuantity

from rspy.units import dimensions

__all__ = ['deg', 'degree', 'degrees', 'rad', 'radian', 'radians', 'decibel', 'dB', 'millihertz', 'mhz', 'mHz',
           'centihertz', 'chz', 'cHz', 'decihertz', 'dhz', 'dHz', 'hertz', 'hz', 'Hz', 'decahertz', 'dahz', 'daHz',
           'hectohertz', 'hhz', 'hHz', 'kilohertz', 'khz', 'kHz', 'megahertz', 'MHz', 'gigahertz', 'ghz', 'GHz',
           'terahertz', 'thz', 'THz', 'petahertz', 'phz', 'PHz', 'nm', 'nanometers', 'nanometer', 'um', 'micrometers',
           'micrometer', 'mm', 'millimeters', 'millimeter', 'cm', 'centimeters', 'centimeter', 'dm', 'decimeters',
           'decimeter', 'm', 'meters', 'meter', 'km', 'kilometers', 'kilometer', 's', 'second', 'seconds', 'minute',
           'minutes', 'h', 'hour', 'hours', 'K', 'kelvins', 'kelvin', 'J', 'joules', 'joule', 'W', 'watt',
           'watts', 'milligram', 'milligrams', 'mg', 'microgram', 'micrograms', 'ug', 'gram', 'grams', 'g',
           'kilogram', 'kilograms', 'kg', 'ampere', 'amperes', 'A', 'One', 'linear', 'celsius', 'fahrenheit']

One = S.One

deg = degree = degrees = sympy_units.degree
rad = radian = radians = sympy_units.radian

decibel = dB = sQuantity("decibel", abbrev="dB")
dB.set_dimension(One)
dB.set_scale_factor(One)

linear = sQuantity("linear")
linear.set_dimension(One)
linear.set_scale_factor(One)

millihertz = mhz = mHz = sQuantity("millihertz", abbrev="mHz")
millihertz.set_dimension(dimensions.frequency)
millihertz.set_scale_factor(1 / 1e3)

centihertz = chz = cHz = sQuantity("centihertz", abbrev="cHz")
centihertz.set_dimension(dimensions.frequency)
centihertz.set_scale_factor(1 / 1e2)

decihertz = dhz = dHz = sQuantity("decihertz", abbrev="dHz")
decihertz.set_dimension(dimensions.frequency)
decihertz.set_scale_factor(1 / 1e1)

hertz = hz = Hz = sQuantity("hertz", abbrev="Hz")
hertz.set_dimension(dimensions.frequency)
hertz.set_scale_factor(One)

decahertz = dahz = daHz = sQuantity("decahertz", abbrev="daHz")
decahertz.set_dimension(dimensions.frequency)
decahertz.set_scale_factor(10)

hectohertz = hhz = hHz = sQuantity("hectohertz", abbrev="hHz")
hectohertz.set_dimension(dimensions.frequency)
hectohertz.set_scale_factor(100)

kilohertz = khz = kHz = sQuantity("kilohertz", abbrev="kHz")
kilohertz.set_dimension(dimensions.frequency)
kilohertz.set_scale_factor(1000)

megahertz = MHz = sQuantity("megahertz", abbrev="MHz")
megahertz.set_dimension(dimensions.frequency)
megahertz.set_scale_factor(1e6)

gigahertz = ghz = GHz = sQuantity("gigahertz", abbrev="GHz")
gigahertz.set_dimension(dimensions.frequency)
gigahertz.set_scale_factor(1e9)

terahertz = thz = THz = sQuantity("terahertz", abbrev="THz")
terahertz.set_dimension(dimensions.frequency)
terahertz.set_scale_factor(1e12)

petahertz = phz = PHz = sQuantity("petahertz", abbrev="PHz")
petahertz.set_dimension(dimensions.frequency)
petahertz.set_scale_factor(1e15)

nm = nanometers = nanometer = sympy_units.nm
um = micrometers = micrometer = sympy_units.um
mm = millimeters = millimeter = sympy_units.mm
cm = centimeters = centimeter = sympy_units.cm
dm = decimeters = decimeter = sympy_units.dm
m = meters = meter = sympy_units.m
km = kilometers = kilometer = sympy_units.km

s = second = seconds = sympy_units.second
minute = minutes = sympy_units.minute
h = hour = hours = sympy_units.hour

K = kelvins = kelvin = sympy_units.K

celsius = sQuantity("celsius")
celsius.set_dimension(dimensions.temperature)
celsius.set_scale_factor(273.15)

fahrenheit = sQuantity("fahrenheit")
fahrenheit.set_dimension(dimensions.temperature)
fahrenheit.set_scale_factor((9./5.))

J = joules = joule = sympy_units.J

W = watt = watts = sympy_units.watt

milligram = milligrams = mg = sympy_units.milligram
microgram = micrograms = ug = sympy_units.microgram
gram = grams = g = sympy_units.gram
kilogram = kilograms = kg = sympy_units.kilogram

ampere = amperes = A = sympy_units.ampere

__unit__ = ['deg', 'degree', 'degrees', 'rad', 'radian', 'radians', 'decibel', 'dB', 'millihertz', 'mhz', 'mHz',
            'centihertz', 'chz', 'cHz', 'decihertz', 'dhz', 'dHz', 'hertz', 'hz', 'Hz', 'decahertz', 'dahz', 'daHz',
            'hectohertz', 'hhz', 'hHz', 'kilohertz', 'khz', 'kHz', 'megahertz', 'MHz', 'gigahertz', 'ghz', 'GHz',
            'terahertz', 'thz', 'THz', 'petahertz', 'phz', 'PHz', 'nm', 'nanometers', 'nanometer', 'um', 'micrometers',
            'micrometer', 'mm', 'millimeters', 'millimeter', 'cm', 'centimeters', 'centimeter', 'dm', 'decimeters',
            'decimeter', 'm', 'meters', 'meter', 'km', 'kilometers', 'kilometer', 's', 'second', 'seconds', 'minute',
            'minutes', 'h', 'hour', 'hours', 'K', 'kelvins', 'kelvin', 'J', 'joules', 'joule', 'W', 'watt',
            'watts', 'milligram', 'milligrams', 'mg', 'microgram', 'micrograms', 'ug', 'gram', 'grams', 'g',
            'kilogram', 'kilograms', 'kg', 'ampere', 'amperes', 'A', 'linear', 'celsius', 'fahrenheit']

__values__ = [deg, degree, degrees, rad, radian, radians, decibel, dB, millihertz, mhz, mHz,
              centihertz, chz, cHz, decihertz, dhz, dHz, hertz, hz, Hz, decahertz, dahz, daHz,
              hectohertz, hhz, hHz, kilohertz, khz, kHz, megahertz, MHz, gigahertz, ghz, GHz,
              terahertz, thz, THz, petahertz, phz, PHz, nm, nanometers, nanometer, um, micrometers,
              micrometer, mm, millimeters, millimeter, cm, centimeters, centimeter, dm, decimeters,
              decimeter, m, meters, meter, km, kilometers, kilometer, s, second, seconds, minute,
              minutes, h, hour, hours, K, kelvins, kelvin, J, joules, joule, W, watt,
              watts, milligram, milligrams, mg, microgram, micrograms, ug, gram, grams, g,
              kilogram, kilograms, kg, ampere, amperes, A, linear, celsius, fahrenheit]



# -*- coding: utf-8 -*-
"""
Created on 03.04.19 by ibaris
"""
from __future__ import division

import pytest

from rspy import Units
from rspy.auxiliary import UnitError
from rspy.units.si_units import *

true_vals = [(gram * nanometer - minute) ** decimeter,
             decimeter,
             decahertz + decimeter - microgram,
             gigahertz + milligram * minute,
             hectohertz / (centimeter * petahertz),
             (ampere + decahertz) / (joule * kelvin),
             ampere,
             centimeter,
             hertz + terahertz,
             radian / millimeter,
             (degree * (kelvin + radian)) ** megahertz,
             kilohertz + second,
             kilometer / minute - millimeter,
             degree * hertz ** milligram,
             hertz * millimeter,
             second ** decahertz,
             gigahertz - kilogram,
             millihertz * (joule + millimeter),
             degree ** 2 + milligram - millihertz,
             decibel + micrometer,
             terahertz - watt,
             nanometer,
             micrometer / terahertz,
             decihertz + hectohertz ** ampere,
             degree * hertz + gigahertz - millimeter,
             -decahertz + meter - second,
             centihertz,
             microgram + micrometer * millihertz,
             decimeter,
             hertz + kelvin,
             kilometer * (-ampere + microgram) / nanometer,
             kilometer * (decahertz - millihertz),
             kilometer,
             -meter + millihertz * (gram - millihertz),
             microgram - terahertz,
             petahertz,
             (gram + kilogram) / nanometer,
             second / hour,
             kilometer * (centimeter ** kilogram - hertz),
             decimeter,
             joule * radian / decihertz,
             decahertz,
             gram ** kilogram - joule,
             kilohertz ** petahertz * second - meter,
             joule,
             centihertz - decimeter,
             (-degree + second) ** hectohertz,
             decihertz,
             minute,
             -megahertz + (meter + watt) ** micrometer,
             decihertz * radian / terahertz - micrometer,
             2 * decimeter * kelvin,
             (centihertz + millihertz) / centimeter,
             (-joule + meter) ** gigahertz,
             -decihertz + gram * (kelvin - microgram),
             (-gram + joule + second) ** gram,
             watt / hertz,
             centihertz ** (second * watt) - centimeter,
             decahertz * (centimeter + watt ** joule),
             (degree ** 2 * hour) ** kilohertz,
             hertz * (centimeter - hectohertz + megahertz),
             (ampere + micrometer) ** hertz,
             microgram ** kilogram,
             ampere + gram + microgram,
             decahertz,
             decimeter - megahertz + meter,
             millimeter * (-gram + micrometer ** decahertz),
             second ** gigahertz,
             (degree * joule + petahertz) ** gram,
             (kilogram / linear) ** (kelvin * watt),
             kilohertz,
             -degree + meter + minute,
             joule * micrometer + micrometer,
             hertz,
             millihertz,
             gigahertz,
             hour * (centimeter - microgram) ** kilohertz,
             minute ** 2,
             millimeter,
             milligram * (ampere - radian) / second,
             linear,
             meter / (decibel * gigahertz * millimeter),
             millihertz,
             joule + millihertz,
             -decibel + nanometer ** centihertz,
             -radian,
             decahertz * (-hour + kilogram ** radian),
             petahertz + (decahertz / micrometer) ** ampere,
             kilogram,
             centimeter + petahertz,
             gigahertz,
             -joule + kilogram,
             gram * micrometer ** millihertz * petahertz,
             microgram * (centihertz + millihertz * millimeter),
             centihertz + microgram / ampere,
             kilogram,
             centihertz * centimeter,
             centihertz * (hour + kelvin + millihertz),
             terahertz,
             nanometer]

str_units = ['grams * nanometers - minutes ** dm',
             'decimeters',
             'decimeters - microgram + dahz',
             'milligrams * minute + ghz',
             'hhz / centimeter / petahertz',
             'daHz + ampere / kelvins / joules ',
             'ampere',
             'cm ',
             'hz + terahertz',
             'rad / mm ',
             'radians + kelvin * deg ** MHz',
             'second + kilohertz ',
             'kilometers / minutes - millimeter',
             'Hz ** milligrams * deg ',
             'millimeters * Hz',
             'second ** decahertz',
             'GHz - kilograms',
             'joules + millimeters * mHz',
             'degree * degrees + mg - mHz',
             'um + decibel',
             'terahertz - watts',
             'nanometer',
             'um / terahertz',
             'hectohertz ** A + dhz',
             'deg * hertz + gigahertz - millimeter',
             'meter - decahertz - second',
             'cHz',
             'micrometer * millihertz + micrograms',
             'decimeters',
             'hertz + kelvins',
             'micrograms - ampere / nm * kilometers',
             'decahertz - mHz * km',
             'km',
             'grams - mHz * mHz - meter',
             'micrograms - terahertz',
             'petahertz',
             'kg + g / nm',
             'seconds / hour',
             'centimeters ** kilograms - hertz * km',
             'decimeter',
             'radian * J / decihertz',
             'decahertz',
             'g ** kilograms - J',
             'kilohertz ** petahertz * seconds - meters',
             'joules',
             'chz - decimeter',
             's - deg ** hHz',
             'dhz',
             'minute',
             'watts + meters ** micrometers - megahertz ',
             'dHz / terahertz * rad - micrometers ',
             'decimeters + dm * K',
             'mHz + chz / centimeters',
             'meter - joule ** gigahertz',
             'kelvin - micrograms * g - dHz',
             's - gram + joule ** g',
             'watts / Hz',
             'cHz ** W ** second - cm',
             'watt ** joule + centimeters * decahertz',
             'degree * h * degree ** kHz',
             'MHz + centimeter - hhz * hertz',
             'micrometer + amperes ** hertz',
             'micrograms ** kg',
             'gram + micrograms + amperes ',
             'daHz',
             'decimeter + meter - megahertz',
             'micrometer ** decahertz - grams * millimeter ',
             'seconds ** ghz ',
             'J * deg + petahertz ** g',
             'kg / linear ** K ** watt ',
             'khz',
             'minute + meter - degree',
             'micrometers * J + micrometer',
             'hz',
             'mHz',
             'GHz',
             'cm - microgram ** kilohertz * hour',
             'minutes * minutes',
             'mm ',
             'ampere - radians * mg / seconds',
             'linear',
             'meter / dB / ghz / millimeters',
             'mHz',
             'mhz + joules',
             'nanometer ** cHz - dB',
             'kilohertz - rad - kilohertz',
             'kilogram ** rad - h * decahertz',
             'dahz / um ** amperes + petahertz',
             'kilograms',
             'petahertz + centimeters',
             'ghz',
             'kilogram - J',
             'um ** millihertz * phz * g',
             'millihertz * millimeter + chz * ug',
             'microgram / ampere + cHz ',
             'kilograms',
             'centihertz * centimeters',
             'kelvin + mhz + hours * chz ',
             'thz',
             'nm']


class TestGetUnit:
    def test_str2unit(self):
        for i, item in enumerate(str_units):
            assert true_vals[i] == Units.get_unit(item)
            assert true_vals[i] == Units.get_unit(true_vals[i])
            assert true_vals[i] == Units.str2unit(item)

    def test_exception(self):
        for i, item in enumerate(str_units):
            with pytest.raises(UnitError):
                item += 'pipi'
                Units.get_unit(item)
                Units.str2unit(item)


class Test2Sting:
    def test_unit2str(self):
        for i, item in enumerate(true_vals):
            assert '[' + str(true_vals[i]) + ']' == Units.unit2str(item)

    def test_unit2sting_none(self):
        for i, item in enumerate(str_units):
            assert '[' + '-' + ']' == Units.unit2str(item)



# -*- coding: utf-8 -*-
"""
Created on 06.04.19 by ibaris
"""
from __future__ import division

import numpy as np

from rspy.ancillary import align_all
from rspy.angles import Angles
from rspy.constants import optical_range
from rspy.sensor.auxiliary import SensorResult
from rspy.units import Units
from rspy.waves import Waves

__all__ = ['Sensor']


class Sensor(Angles, Waves):
    """
    A class to build a sensor with sensing geometry and frequency.

    See Also
    --------
    rspy.angles.Angles
    rspy.waves.Waves
    """

    def __init__(self, value, iza, vza, raa=None, iaa=None, vaa=None, normalize=False, nbar=0.0,
                 angle_unit='DEG', dtype=np.double, unit='GHz', output='cm', name=None):

        """

        Parameters
        ----------
        value : int, float, np.ndarray, respy.units.quantity.Quantity
            Frequency or wavelength.
        iza, vza, raa, iaa, vaa : int, float or array_like
            Incidence (iza) and scattering (vza) zenith angle, relative azimuth (raa) angle, incidence and viewing
            azimuth angle (ira, vra). If raa is defined, ira and vra are not mandatory.
        alpha, beta: int, float or array_like
            The Euler angles of the particle orientation (degrees).
        normalize : boolean, optional
            Set to 'True' to make kernels 0 at nadir view illumination. Since all implemented kernels are normalized
            the default value is False.
        nbar : float, optional
            The sun or incidence zenith angle at which the isotropic term is set
            to if normalize is True. The default value is 0.0.
        angle_unit : {'DEG', 'RAD', 'deg', 'rad'}, optional
            * 'DEG': All input angles (iza, vza, raa) are in [DEG] (default).
            * 'RAD': All input angles (iza, vza, raa) are in [RAD].
        dtype : data-type
            Desired data type of all values. Default is np.double.
        unit : str, respy.units.Units, sympy.physics.units.quantities.Quantity
            Unit of input. Default is 'GHz'.
        output : str, respy.units.Units, sympy.physics.units.quantities.Quantity
            Unit of output. Default is 'cm'.
        name : str
            A name for the created sensor.
        """
        if raa is None:
            iza, vza, iaa, vaa, value = align_all((iza, vza, iaa, vaa, value))

        else:
            iza, vza, raa, value = align_all((iza, vza, raa, value))

        Angles.__init__(self, iza=iza, vza=vza, raa=raa, iaa=iaa, vaa=vaa, normalize=normalize, nbar=nbar,
                        angle_unit=angle_unit, align=True, dtype=dtype)

        if normalize:
            value = np.append(value, value[-1])

        Waves.__init__(self, value=value, unit=unit, output=output)

        self.name = name
        self.input_unit = unit
        self.output = output

        self.args = (value, iza, vza, raa, iaa, vaa, normalize, nbar, angle_unit, dtype, unit, output, name)

        self.__band_names = ['Band 1', 'Band 2', 'Band 3', 'Band 4', 'Band 5', 'Band 6', 'Band 7', 'Band 8', 'Band 9']
        self.__abbrev = ['B1', 'B2', 'B3', 'B4', 'B5', 'B6', 'B7', 'B8', 'B9']

    def __repr__(self):
        prefix = '<{0} '.format(self.__class__.__name__)
        sep = ', '
        if self.input_unit in Units.frequency.keys():
            arrstr = np.array2string(self.frequency,
                                     separator=sep,
                                     prefix=prefix)

            unit = self.frequency_unit

        else:
            arrstr = np.array2string(self.wavelength,
                                     separator=sep,
                                     prefix=prefix)

            unit = self.wavelength_unit

        if self.name is None or self.name == b'':
            return '{0}{1} [{2}]>'.format(prefix, arrstr, unit)

        return '{0}{1} {2} in [{3}]>'.format(prefix, arrstr, self.name, unit)

    # --------------------------------------------------------------------------------------------------------
    # Callable Methods
    # --------------------------------------------------------------------------------------------------------
    def align_with(self, value):
        """
        Align all angles with another array.

        Parameters
        ----------
        value : array_like

        Returns
        -------
        value : array_like
            Align value.

        Note
        ----
        If len(value) > Angles.shape[1] then the angles inside Angles class will be aligned and it has no effect on
        value. If len(value) < Angles.shape[1] the output of value will be have the same len as Angles and it has no
        effect on the angles within the Angles class.
        """
        # Align Angles -----------------------------------------------------------------------------------------------
        data = [item for item in self.array]

        if isinstance(value, (tuple, list)):
            data = tuple(value) + tuple(data, )
        else:
            data = (value,) + tuple(data, )

        data = align_all(data)

        # DEG Angles
        dataDeg = [item for item in self.arrayDeg]

        if isinstance(value, (tuple, list)):
            dataDeg = tuple(value) + tuple(dataDeg, )
        else:
            dataDeg = (value,) + tuple(dataDeg, )

        dataDeg = align_all(dataDeg)

        self.array = np.asarray(data[-7:])
        self.arrayDeg = np.asarray(dataDeg[-7:])

        # Align Frequencies ------------------------------------------------------------------------------------------
        data = [item for item in self.values]

        if isinstance(value, (tuple, list)):
            data = tuple(value) + tuple(data, )
        else:
            data = (value,) + tuple(data, )

        data = align_all(data)

        self.values = np.asarray(data[-3:])

        return data[0:-3]

    def L8(self, value):
        """ Store Landsat 8 bands.
        Store reflectance for Landsat 8 bands
        Returns
        -------
        Landsat bands B2 - B7: numpy.ndarray
            The mean values of Landsat 8 Bands from numpy.array([B2 - B7]). See `Notes`.
        Note
        ----
        Landsat 8 Bands are in [nm]:
            * B2 = 452 - 512
            * B3 = 533 - 590
            * B4 = 636 - 673
            * B5 = 851 - 879
            * B6 = 1566 - 1651
            * B7 = 2107 - 2294
        """

        if len(optical_range) != len(value):
            raise AssertionError(
                "Value must contain continuous reflectance values from from 400 until 2500 nm with a length of "
                "2101. The actual length of value is {0}".format(str(len(value))))

        b2 = (452, 452 + 60)
        b3 = (533, 533 + 57)
        b4 = (636, 636 + 37)
        b5 = (851, 851 + 28)
        b6 = (1566, 1566 + 85)
        b7 = (2107, 2107 + 187)

        array = np.array([optical_range, value]).transpose()

        LRefB2 = array[(array[:, 0] >= b2[0]) & (array[:, 0] <= b2[1])]
        LRefB3 = array[(array[:, 0] >= b3[0]) & (array[:, 0] <= b3[1])]
        LRefB4 = array[(array[:, 0] >= b4[0]) & (array[:, 0] <= b4[1])]
        LRefB5 = array[(array[:, 0] >= b5[0]) & (array[:, 0] <= b5[1])]
        LRefB6 = array[(array[:, 0] >= b6[0]) & (array[:, 0] <= b6[1])]
        LRefB7 = array[(array[:, 0] >= b7[0]) & (array[:, 0] <= b7[1])]

        bands = np.array([LRefB2[:, 1].mean(), LRefB3[:, 1].mean(), LRefB4[:, 1].mean(), LRefB5[:, 1].mean(),
                          LRefB6[:, 1].mean(), LRefB7[:, 1].mean()])

        BAND_NAMES = self.__band_names[1:-2]
        abbrev = self.__abbrev[1:-2]
        # spectra = ['Blue', 'Green', 'Red', 'NIR', 'SWIR 1', 'SWIR 2']

        # names = list()
        # if self.name is None:
        #     for i, item in enumerate(BAND_NAMES):
        #         names.append(SENSOR + ' ' + item + ' ' + '(' + spectra[i] + ')')
        #
        # else:
        #     for i, item in enumerate(BAND_NAMES):
        #         names.append(self.name + ' ' + SENSOR + ' ' + item + ' ' + '(' + spectra[i] + ')')

        L8 = SensorResult(name='LANDSAT 8')

        for i, item in enumerate(BAND_NAMES):
            L8[abbrev[i]] = bands[i]

        return L8

    def Aster(self, value):
        """ Store Aster bands.
        Store reflectance for Aster bands
        Returns
        -------
        Aster bands B1 - B9: numpy.ndarray
            The mean values of Aster Bands from B1 - B9. See `Notes`.
        Note
        ----
        Landsat 8 Bands are in [nm]:
            * B1 = 520 - 600
            * B2 = 630 - 690
            * B3 = 760 - 860
            * B4 = 1600 - 1700
            * B5 = 2145 - 2185
            * B6 = 2185 - 2225
            * B7 = 2235 - 2285
            * B8 = 2295 - 2365
            * B9 = 2360 - 2430
        """
        if len(optical_range) != len(value):
            raise AssertionError(
                "Value must contain continuous reflectance values from from 400 until 2500 nm with a length of "
                "2101. The actual length of value is {0}".format(str(len(value))))

        array = np.array([optical_range, value]).transpose()

        b1 = (520, 600)
        b2 = (630, 690)
        b3 = (760, 860)
        b4 = (1600, 1700)
        b5 = (2145, 2185)
        b6 = (2185, 2225)
        b7 = (2235, 2285)
        b8 = (2295, 2365)
        b9 = (2360, 2430)

        ARefB1 = array[(array[:, 0] >= b1[0]) & (array[:, 0] <= b1[1])]
        ARefB2 = array[(array[:, 0] >= b2[0]) & (array[:, 0] <= b2[1])]
        ARefB3 = array[(array[:, 0] >= b3[0]) & (array[:, 0] <= b3[1])]
        ARefB4 = array[(array[:, 0] >= b4[0]) & (array[:, 0] <= b4[1])]
        ARefB5 = array[(array[:, 0] >= b5[0]) & (array[:, 0] <= b5[1])]
        ARefB6 = array[(array[:, 0] >= b6[0]) & (array[:, 0] <= b6[1])]
        ARefB7 = array[(array[:, 0] >= b7[0]) & (array[:, 0] <= b7[1])]
        ARefB8 = array[(array[:, 0] >= b8[0]) & (array[:, 0] <= b8[1])]
        ARefB9 = array[(array[:, 0] >= b9[0]) & (array[:, 0] <= b9[1])]

        bands = np.array([ARefB1[:, 1].mean(), ARefB2[:, 1].mean(), ARefB3[:, 1].mean(), ARefB4[:, 1].mean(),
                          ARefB5[:, 1].mean(), ARefB6[:, 1].mean(), ARefB7[:, 1].mean(), ARefB8[:, 1].mean(),
                          ARefB9[:, 1].mean()])

        BAND_NAMES = self.__band_names
        abbrev = self.__abbrev

        # names = list()
        # if self.name == b'':
        #     for i, item in enumerate(BAND_NAMES):
        #         names.append(SENSOR + ' ' + item + ' ' + '(' + spectra[i] + ')')
        #
        # else:
        #     for i, item in enumerate(BAND_NAMES):
        #         names.append(self.name + ' ' + SENSOR + ' ' + item + ' ' + '(' + spectra[i] + ')')

        ASTER = SensorResult(name='ASTER')

        for i, item in enumerate(BAND_NAMES):
            ASTER[abbrev[i]] = bands[i]

        return ASTER

    def ndvi(self, sensor=None, red=None, nir=None):
        """
        Get the NDVI value.

        Parameters
        ----------
        sensor : dict, optional
            A dictionary with keys:
                * 'name': {'LANDSAT 8', 'ASTER'}
                * 'B2 - B7': Values for Band 2 until Band 7 if 'name' is 'LANDSAT 8'
                * 'B1 - B9': Values for Band 1 until Band 9 if 'name' is 'ASTER'
        red : numpy.ndarray, optional
            Define a custom value for the red band.
        nir : numpy.ndarray, optional
            Define a custom value for the nir band.

        Returns
        -------
        out : float
            NDVI value for setup.
        """

        if sensor is None and red is None and nir is None:
            raise ValueError("Parameter `sensor` or `red` and `nir` must be defined.")

        if sensor is not None:
            if isinstance(sensor, dict):
                if hasattr(sensor, 'name'):
                    if sensor['name'] == 'LANDSAT 8':
                        red = sensor['B4']
                        nir = sensor['B5']
                    elif sensor['name'] == 'ASTER':
                        red = sensor['B2']
                        nir = sensor['B3']
                    else:
                        raise AssertionError("Sensor {0} not supported. Only 'LANDSAT 8' or 'ASTER' are supported.")
                else:
                    raise AssertionError("If parameter sensor is defined it must be contain a key with `name` with "
                                         "value 'LANDSAT 8' or 'ASTER'.")

            else:
                raise TypeError("Parameter sensor must be a dictionary.")

        ndvi = (nir - red) / (nir + red)

        return ndvi

    def sr(self, sensor=None, red=None, nir=None):
        """
        Get the Red, NIR Ratio.

        Parameters
        ----------
        sensor : dict, optional
            A dictionary with keys:
                * 'name': {'LANDSAT 8', 'ASTER'}
                * 'B2 - B7': Values for Band 2 until Band 7 if 'name' is 'LANDSAT 8'
                * 'B1 - B9': Values for Band 1 until Band 9 if 'name' is 'ASTER'
        red : numpy.ndarray, optional
            Define a custom value for the red band.
        nir : numpy.ndarray, optional
            Define a custom value for the nir band.

        Returns
        -------
        out : float
            SR value for setup.
        """

        if sensor is None and red is None and nir is None:
            raise ValueError("Parameter `sensor` or `red` and `nir` must be defined.")

        if sensor is not None:
            if isinstance(sensor, dict):
                if hasattr(sensor, 'name'):
                    if sensor['name'] == 'LANDSAT 8':
                        red = sensor['B4']
                        nir = sensor['B5']
                    elif sensor['name'] == 'ASTER':
                        red = sensor['B2']
                        nir = sensor['B3']
                    else:
                        raise AssertionError("Sensor {0} not supported. Only 'LANDSAT 8' or 'ASTER' are supported.")
                else:
                    raise AssertionError("If parameter sensor is defined it must be contain a key with `name` with "
                                         "value 'LANDSAT 8' or 'ASTER'.")

            else:
                raise TypeError("Parameter sensor must be a dictionary.")

        sr = nir / red

        return sr



# -*- coding: utf-8 -*-
"""
Created on 06.04.19 by ibaris
"""
from __future__ import division

import numpy as np

import rspy.constants as const
from rspy.units import Units
from rspy.auxiliary import UnitError
from rspy.ancillary import align_all

__all__ = ['Waves']


class Waves(object):
    """
    This is a simple transverse wave travelling in a one-dimensional space.

    Attributes
    ----------
    Waves.band : str
        Shows the band that the input belongs to.
    Waves.region : str
        Shows the region that the input belongs to.
    Waves.frequency : `Quantity` object
        Frequency.
    Waves.wavelength : `Quantity` object
        Wavelength.
    Waves.len : int
        Length of array.
    Waves.shape : tuple
        Shape of array.
    Waves.value : np.ndarray
        An array with the values of frequency, wavelength and wavenumber.
    Waves.wavenumber : `Quantity` object
        Wavenumber.
    Waves.speed : `Quantity` object
        Speed of the electromagnetic wave.
    Waves.angular_speed : `Quantity` object
        Angular velocity of the electromagnetic wave.

    Methods
    -------
    align_with(value)
        Expand all input values to the same length depend on an external array.
    compute_frequency(wavelength, unit, output, quantity=True)
        Static method to convert wavelengths in frequencies.
    compute_wavelength(frequency, unit, output, quantity=True)
        Static method to convert frequencies in wavelengths.
    compute_wavenumber(frequency, unit, output, quantity=True)
        Static method to convert frequencies in free space wavenumbers.
    planck(tempreture, wavelength=True)
        Evaluate Planck's radiation law.
    """

    def __init__(self, value, unit='GHz', output='cm'):
        """
        A class to describe electromagnetic waves.

        Parameters
        ----------
        value : int, float, np.ndarray, respy.units.quantity.Quantity
            Frequency or wavelength.
        unit : str, respy.units.Units, sympy.physics.units.quantities.Quantity
            Unit of input. Default is 'GHz'.
        output : str, respy.units.Units, sympy.physics.units.quantities.Quantity
            Unit of output. Default is 'cm'.

        """

        # Prepare Input Data and set values ----------------------------------------------------------------------------
        value = np.asarray(value).flatten()

        # Self Definitions ---------------------------------------------------------------------------------------------
        self.__unit = Units.get_unit(unit)
        self.__output = Units.get_unit(output)

        # Assign Input Parameter ---------------------------------------------------------------------------------------
        if self.__unit in Units.frequency.values():
            self.__frequency_unit = self.__unit

            if self.__output not in Units.length.values():
                raise UnitError("Output unit {} is not a valid unit if value is a frequency.".format(str(output)))
            else:
                self.__wavelength_unit = self.__output

            self.__frequency = np.asarray(value).flatten()

            self.__wavelength = Waves.compute_wavelength(self.__frequency, self.__frequency_unit,
                                                         output=self.__wavelength_unit)

        elif self.__unit in Units.length.values():
            self.__wavelength_unit = self.__unit

            if self.__output not in Units.frequency.values():
                raise UnitError("Output unit {} is not a valid unit if value is a wavelength.".format(str(output)))
            else:
                self.__frequency_unit = self.__output

            self.__wavelength = np.asarray(value).flatten()
            self.__frequency = Waves.compute_frequency(self.__wavelength, unit=self.__wavelength_unit,
                                                       output=self.__frequency_unit)

        else:
            raise UnitError("Input must be a frequency or a wavelength. "
                            "If value is a frequency, unit must be {0}. "
                            "When entering a wavelength, unit must be {1}.".format(str(Units.frequency.keys()),
                                                                                   str(Units.length.keys())))

        # Additional Calculation ---------------------------------------------------------------------------------------
        self.__wavenumber = Waves.compute_wavenumber(self.__frequency, self.__frequency_unit,
                                                     output=self.__wavelength_unit)

        self.values = np.array([self.__frequency, self.__wavelength, self.__wavenumber])

        self.__region = None
        self.__band = None

    # --------------------------------------------------------------------------------------------------------
    # Magic Methods
    # --------------------------------------------------------------------------------------------------------
    def __repr__(self):
        prefix = '<{0} '.format(self.__class__.__name__)
        sep = ','
        arrstr_freq = np.array2string(self.frequency,
                                      separator=sep,
                                      prefix=prefix)

        arrstr_wave = np.array2string(self.wavelength,
                                      separator=sep,
                                      prefix=prefix)

        arrstr_wavenumber = np.array2string(self.wavenumber,
                                            separator=sep,
                                            prefix=prefix)

        wavenumber = '{0}{1} Wavenumber in free space [{2}]>'.format(prefix, arrstr_wavenumber,
                                                                     Units.unit2str(self.wavenumber_unit))

        if self.__band is None:
            freq = '{0}{1} Frequency in region {2} in [{3}]>'.format(prefix, arrstr_freq, self.__region,
                                                                     Units.unit2str(self.frequency_unit))
            wave = '{0}{1} Wavelength in region {2} in [{3}]>'.format(prefix, arrstr_wave, self.__region,
                                                                      Units.unit2str(self.wavelength_unit))

        else:
            freq = '{0}{1} Frequency in region {2} ({3}-Band) in [{4}]>'.format(prefix, arrstr_freq, self.__region,
                                                                                self.__band,
                                                                                Units.unit2str(self.frequency_unit))
            wave = '{0}{1} Wavelength in region {2} ({3}-Band) in [{4}]>'.format(prefix, arrstr_wave, self.__region,
                                                                                 self.__band,
                                                                                 Units.unit2str(self.wavelength_unit))

        return freq + '\n' + wave + '\n' + wavenumber

    def __len__(self):
        return len(self.__frequency)

    # --------------------------------------------------------------------------------------------------------
    # Properties
    # --------------------------------------------------------------------------------------------------------
    @property
    def len(self):
        """
        Length of array

        Returns
        -------
        len : int
        """
        return len(self.frequency)

    @property
    def shape(self):
        """
        Shape of array

        Returns
        -------
        shape : tuple
        """
        return self.frequency.shape

    @property
    def band(self):
        return self.__band

    @property
    def region(self):
        return self.__region

    @property
    def frequency(self):
        return self.values[0,]

    @property
    def frequency_unit(self):
        return self.__frequency_unit

    @property
    def wavelength(self):
        return self.values[1,]

    @property
    def wavelength_unit(self):
        return self.__wavelength_unit

    @property
    def wavenumber(self):
        return self.values[2,]

    @property
    def wavenumber_unit(self):
        return 1 / self.__wavelength_unit

    # --------------------------------------------------------------------------------------------------------
    # Callable Methods
    # --------------------------------------------------------------------------------------------------------
    def align_with(self, value):
        """
        Align all angles with another array.

        Parameters
        ----------
        value : array_like

        Returns
        -------
        value : array_like
            Align value.

        Note
        ----
        If len(value) > EMW.shape[1] then the angles inside Angles class will be aligned and it has no effect on
        value. If len(value) < EMW.shape[1] the output of value will be have the same len as Angles and it has no
        effect on the angles within the Angles class.
        """
        # RAD Angles

        data = [item for item in self.values]

        if isinstance(value, (tuple, list)):
            data = tuple(value) + tuple(data, )
        else:
            data = (value,) + tuple(data, )

        data = align_all(data)

        self.values = np.asarray(data[-3:])

        return data[0:-3]

    @staticmethod
    def compute_frequency(wavelength, unit='cm', output='GHz'):
        """
        Convert wavelengths in frequencies.

        Parameters
        ----------
        wavelength : int, float, np.ndarray, object
            Wavelength as int, float, numpy.ndarray or as a respy.units.quantity.Quantity object.
        unit : str, object
            Unit of the wavelength (default is 'cm'). See respy.units.Units.length.keys() for available units.
            This is optional if the input is an respy.units.quantity.Quantity object.
        output : str, object
            Unit of entered frequency (default is 'GHz'). See respy.units.Units.frequency.keys() for available units.

        Returns
        -------
        frequency: float, np.ndarray or respy.units.quantity.Quantity

        """
        unit = Units.get_unit(unit)
        output = Units.get_unit(output)

        c = Units.convert_to(const.c, 'm / s', unit / Units.time.s)
        f = c / wavelength

        return Units.convert_to(f, 'Hz', output)

    @staticmethod
    def compute_wavelength(frequency, unit='GHz', output='cm'):
        """
        Convert frequencies in wavelength.

        Parameters
        ----------
        frequency : int, float, np.ndarray, object
            Frequency as int, float, numpy.ndarray or as a respy.units.quantity.Quantity object.
        unit : str, object
            Unit of entered frequency (default is 'GHz'). See respy.units.Units.frequency.keys() for available units.
            This is optional if the input is an respy.units.quantity.Quantity object.
        output : str, object
            Unit of the wavelength (default is 'cm'). See respy.units.Units.length.keys() for available units.

        Returns
        -------
        Wavelength: float, np.ndarray or respy.units.quantity.Quantity
        """
        unit = Units.get_unit(unit)
        output = Units.get_unit(output)

        frequency = Units.convert_to(frequency, unit, '1 / s')

        w = const.c / frequency

        return Units.convert_to(w, 'm', output)

    @staticmethod
    def compute_wavenumber(frequency, unit='GHz', output='cm'):
        """
        Convert frequencies in free space wavenumbers.

        Parameters
        ----------
        frequency : int, float, np.ndarray, object
            Frequency as int, float, numpy.ndarray or as a respy.units.quantity.Quantity object.
        unit : str, object
            Unit of entered frequency (default is 'GHz'). See respy.units.Units.frequency.keys() for available units.
            This is optional if the input is an respy.units.quantity.Quantity object.
        output : str, object
            Unit of the wavelength (default is 'cm'). See respy.units.Units.length.keys() for available units.

        Returns
        -------
        wavenumber: float, np.ndarray or respy.units.quantity.Quantity
        """
        return 2 * const.pi / Waves.compute_wavelength(frequency, unit=unit, output=output)



# -*- coding: utf-8 -*-
"""
Created on 06.04.19 by ibaris
"""
from __future__ import division

import numpy as np
import pytest

from rspy import Waves, Units, UnitError, Sensor

n = 50

n1 = np.random.randint(60, 101)

freq = np.random.uniform(0.5, 100000000, n1)  # Frequencies in Hz
lam = np.random.uniform(0.01, 1000000, n1)  # Frequencies in cm

n2 = np.random.randint(120, 150)
n3 = np.random.randint(30, 50)

random_array = np.random.uniform(1, np.pi / 2, n2)
random_array2 = np.random.uniform(1, np.pi / 2, n3)

iza, vza, iaa, vaa, nbar = np.random.uniform(1, np.pi, n1), np.random.uniform(1, np.pi, n1), \
                           np.random.uniform(1, 2 * np.pi, n1), np.random.uniform(1, 2 * np.pi, n1), \
                           np.random.uniform(1, np.pi / 4, 1)

izaDeg, vzaDeg, iaaDeg, vaaDeg, nbarDeg = np.rad2deg(iza), np.rad2deg(vza), np.rad2deg(iaa), np.rad2deg(
    vaa), np.rad2deg(nbar)

raa = iaa - vaa
raaDeg = iaaDeg - vaaDeg


class TestInit:
    def test_freq(self):
        for x in range(n):
            units = Units['frequency']
            ind = np.random.choice(len(units.values()))
            unit_values = list(units.values())
            unit1 = unit_values[ind]

            units = Units['length']
            ind = np.random.choice(len(units.values()))
            unit_values = list(units.values())
            unit2 = unit_values[ind]

            w = Sensor(value=freq, iza=izaDeg, vza=vzaDeg, iaa=iaaDeg, vaa=vaaDeg,
                       normalize=False, nbar=nbarDeg, unit=unit1, output=unit2)

            assert np.allclose(w.frequency, freq)
            assert w.frequency_unit == unit1
            assert w.wavelength_unit == unit2
            assert w.wavenumber_unit == 1 / unit2

    def test_lam(self):
        for x in range(n):
            units = Units['frequency']
            ind = np.random.choice(len(units.values()))
            unit_values = list(units.values())
            unit1 = unit_values[ind]

            units = Units['length']
            ind = np.random.choice(len(units.values()))
            unit_values = list(units.values())
            unit2 = unit_values[ind]

            w = Sensor(value=lam, iza=izaDeg, vza=vzaDeg, iaa=iaaDeg, vaa=vaaDeg,
                       normalize=False, nbar=nbarDeg, unit=unit2, output=unit1)

            assert np.allclose(w.wavelength, lam)
            assert w.frequency_unit == unit1
            assert w.wavelength_unit == unit2
            assert w.wavenumber_unit == 1 / unit2

    def test_utility(self):
        for x in range(n):
            units = Units['frequency']
            ind = np.random.choice(len(units.values()))
            unit_values = list(units.values())
            unit1 = unit_values[ind]

            units = Units['length']
            ind = np.random.choice(len(units.values()))
            unit_values = list(units.values())
            unit2 = unit_values[ind]

            w = Sensor(value=freq, iza=izaDeg, vza=vzaDeg, iaa=iaaDeg, vaa=vaaDeg,
                       normalize=False, nbar=nbarDeg, unit=unit1, output=unit2)

            assert len(w) == len(freq)
            assert w.len == len(freq)



# -*- coding: utf-8 -*-
"""
Created on 16.04.19 by ibaris
"""
from __future__ import division

import numpy as np
import pytest

from rspy.ancillary import *
from rspy.intensity import *

n = 50


class TestErrors:
    def test1(self):
        for x in range(n):
            BRDF = np.random.uniform(0.00001, 0.1, n)
            with pytest.raises(ValueError):
                I = Intensity(BRDF, value_unit="BRDF", angle_unit='pipimumu')

    def test2(self):
        for x in range(n):
            BRDF = np.random.uniform(0.00001, 0.1, n)
            I = Intensity(BRDF, value_unit="BRDF")
            with pytest.raises(KeyError):
                I = I['KAKA']



# -*- coding: utf-8 -*-
"""
Created on 01.04.2019 by Ismail Baris
"""
from __future__ import division

import numpy as np

from rspy import Angles
from rspy.auxiliary.types import __DTYPES__

n = 100

random_dtype = np.random.choice(__DTYPES__, 1)[0]

iza, vza, iaa, vaa, alpha, beta, nbar = np.random.uniform(1, np.pi / 2, n), np.random.uniform(1, np.pi / 2, n), \
                                        np.random.uniform(1, 2 * np.pi, n), np.random.uniform(1, 2 * np.pi, n), \
                                        np.random.uniform(1, np.pi, n), np.random.uniform(1, 2 * np.pi, n), \
                                        np.random.uniform(1, np.pi / 4, 1)

izaDeg, vzaDeg, iaaDeg, vaaDeg, alphaDeg, betaDeg, nbarDeg = np.rad2deg(iza), np.rad2deg(vza), np.rad2deg(
    iaa), np.rad2deg(vaa), np.rad2deg(alpha), np.rad2deg(beta), np.rad2deg(nbar)

raa = iaa - vaa
raaDeg = iaaDeg - vaaDeg

degitems = [izaDeg, vzaDeg, iaaDeg, vaaDeg, raaDeg, alphaDeg, betaDeg]
raditems = [iza, vza, iaa, vaa, raa, alpha, beta]

degitems_key = ["izaDeg", "vzaDeg", "iaaDeg", "vaaDeg", "raaDeg", "alphaDeg", "betaDeg"]
raditems_key = ["iza", "vza", "iaa", "vaa", "raa", "alpha", "beta"]

property_list = ["len", "shape", "B", "BDeg", "mui", "muv", "phi", "dtype", "nbar", "nbarDeg", "normalize"]

# "geometries", "geometriesDeg", "array", "arraDeg"

property_list_values = [len(iza), (7, len(iza)), (1 / np.cos(iza) + 1 / np.cos(vza)),
                        (1 / np.cos(izaDeg) + 1 / np.cos(vzaDeg)),
                        np.cos(iza), np.cos(vza), np.abs((raa % (2. * np.pi))), random_dtype, nbar, nbarDeg, False]

array = np.array([iza, vza, raa, iaa, vaa, alpha, beta])
arrayDeg = np.array([izaDeg, vzaDeg, raaDeg, iaaDeg, vaaDeg, alphaDeg, betaDeg])


class TestProperties:
    def test_properties_and_getitem_conversion_DEG(self):
        align = True
        angle_unit = 'DEG'

        ang = Angles(iza=izaDeg, vza=vzaDeg, iaa=iaaDeg, vaa=vaaDeg, alpha=alphaDeg, beta=betaDeg,
                     normalize=False, nbar=nbarDeg, angle_unit=angle_unit, align=align, dtype=random_dtype)

        for i, item in enumerate(degitems):
            assert np.allclose(ang[degitems_key[i]], item)

        for i, item in enumerate(raditems):
            assert np.allclose(ang[raditems_key[i]], item)

        for i, item in enumerate(property_list_values):
            try:
                assert np.allclose(ang[property_list[i]], item)
            except (ValueError, TypeError):
                assert ang[property_list[i]] == item

        assert np.allclose(ang.array, array)
        assert np.allclose(ang.arrayDeg, arrayDeg)
        assert ang.align == align
        assert ang.angle_unit == angle_unit

        raditems_geom = [iza, vza, raa, iaa, vaa, alpha, beta]
        degitems_geom = [izaDeg, vzaDeg, raaDeg, iaaDeg, vaaDeg, alphaDeg, betaDeg]

        for i in range(len(iza)):
            radgeom = ang.geometries[i]
            degdgeom = ang.geometriesDeg[i]

            for j, _ in enumerate(radgeom):
                assert np.allclose(radgeom[j], raditems_geom[j][i])
                assert np.allclose(degdgeom[j], degitems_geom[j][i])

    def test_properties_and_getitem_conversion_RAD(self):
        align = True
        angle_unit = 'RAD'

        ang = Angles(iza=iza, vza=vza, iaa=iaa, vaa=vaa, alpha=alpha, beta=beta,
                     normalize=False, nbar=nbar, angle_unit=angle_unit, align=align, dtype=random_dtype)

        for i, item in enumerate(degitems):
            assert np.allclose(ang[degitems_key[i]], item)

        for i, item in enumerate(raditems):
            assert np.allclose(ang[raditems_key[i]], item)

        for i, item in enumerate(property_list_values):
            try:
                assert np.allclose(ang[property_list[i]], item)
            except (ValueError, TypeError):
                assert ang[property_list[i]] == item

        assert np.allclose(ang.array, array)
        assert np.allclose(ang.arrayDeg, arrayDeg)
        assert ang.align == align
        assert ang.angle_unit == angle_unit

        raditems_geom = [iza, vza, raa, iaa, vaa, alpha, beta]
        degitems_geom = [izaDeg, vzaDeg, raaDeg, iaaDeg, vaaDeg, alphaDeg, betaDeg]

        for i in range(len(iza)):
            radgeom = ang.geometries[i]
            degdgeom = ang.geometriesDeg[i]

            for j, _ in enumerate(radgeom):
                assert np.allclose(radgeom[j], raditems_geom[j][i])
                assert np.allclose(degdgeom[j], degitems_geom[j][i])



# -*- coding: utf-8 -*-
"""
Created on  by Ismail Baris
"""
from __future__ import division

from numpy import cos, tan

from rspy.constants import pi

__all__ = ['d2r', 'r2d', 'sec', 'cot']


def d2r(angle):
    """
    Convert degrees to radians.
    Parameter:
    ----------
    angle : int, float or numpy.ndarray
        Angle in [DEG].
    """

    return angle * pi / 180.0


def r2d(angle):
    """
    Convert radians to degree.
    Parameter:
    ----------
    angle : int, float or numpy.ndarray
        Angle in [RAD].
    """

    return angle * 180. / pi


def sec(angle):
    """
    Secant of an angle.
    """
    return 1 / cos(angle)


def cot(x):
    """
    Cotangent of an angle.
    """
    return 1 / tan(x)



# -*- coding: utf-8 -*-
# cython: cdivision=True, boundscheck=False, wraparound=False, nonecheck=False
"""
Created on 03.04.19 by ibaris
"""
from __future__ import division
import numpy as np
cimport numpy as np
from sympy.physics.units import convert_to as sympy_convert_to
from sympy import S
from rspy.bin.bin_units.dtypes cimport DTYPE_ARRAY


cdef double[:] bin_sym_convert_to(DTYPE_ARRAY expr, object unit):
    """
    Convert between units via `sympy.convert_to`.
    
    Parameters
    ----------
    expr : numpy.ndarray
        An array with sympy unit expressions.
    unit :  sympy.core.mul.Mul, sympy.physics.units.quantities.Quantity
        The unit to which you want to convert.

    Returns
    -------
    double[:]
    
    """
    cdef:
        Py_ssize_t i, x = expr.shape[0]

    cdef double[:] value = np.empty_like(expr, dtype=np.double)

    for i in range(x):
        arg = sympy_convert_to(expr[i], unit).n()

        if arg.args[0] != S.Zero:
            value[i] = arg.args[0]
        else:
            value[i] = arg

    return value



# -*- coding: utf-8 -*-
"""
Created on 01.04.2019 by Ismail Baris
"""
from __future__ import division

from sympy.physics.units.dimensions import (Dimension, frequency, length, energy, power,
                                            temperature, volume, time, mass, current)

__all__ = ['angle', 'area', 'volume', 'frequency', 'length', 'energy', 'power', 'temperature', 'time', 'mass',
           'current']

angle = Dimension(name='angle')
area = Dimension(name='area', symbol='A')
volume = volume
frequency = frequency
length = length
energy = energy
power = power
temperature = temperature
time = time
mass = mass
current = current



# -*- coding: utf-8 -*-
"""
Created on 01.04.2019 by Ismail Baris
"""
from __future__ import division

import pytest

from rspy import Units

dir_list = ['_Units__unit_dict',
            'angle',
            'area',
            'current',
            'dimensions',
            'energy',
            'frequency',
            'length',
            'mass',
            'other',
            'power',
            'temperature',
            'time',
            'units',
            'volume']


class TestDictMethods:
    def test_dicts(self):
        assert dir(Units) == dir_list

    def test_dicts_err(self):
        keys = ['a', 'b', 'c']
        for item in keys:
            with pytest.raises(AttributeError):
                first = Units.item

            with pytest.raises(KeyError):
                second = Units[item]



    def test_links_to_dimensions(self):
        for i in range(1, len(dir_list)):
            item = dir_list[i]

            if item in ['dimensions', 'units']:
                pass
            else:
                assert item in str(Units[item].__class__).lower()


class TestAttributes:
    def test_dim(self):
        assert (len(Units.dimensions) == len(Units.__unit_dict__) - 1)

    def test_len_setup(self):
        lens = 0

        for item in Units:
            lens += len(item)

        lens -= len(Units.dimensions)  # Unit class has 11 dimensions, which contributes to length.

        assert len(Units.units) == lens + 3



# -*- coding: utf-8 -*-
"""
Created on 01.04.2019 by Ismail Baris
"""
from __future__ import division

import numpy as np
import pytest

from rspy.units.auxiliary import *

dicts = [Frequency, Length, Energy, Power, Time, Temperature, Mass, Current, Other, Area, Volume,
         Angle]


class TestAllDicts:
    def test_dicts(self):
        for item in dicts:
            a = np.random.random()

            a_item = item(a=a)

            assert a_item.a == a

    def test_dicts_err(self):
        for item in dicts:
            a = np.random.random()

            a_item = item(a=a)

            with pytest.raises(AttributeError):
                a_item.b

    def test_dicts_dir(self):
        for item in dicts:
            a = np.random.random()

            a_item = item(a=a)

            assert dir(a_item) == ['a']

    def test_repr(self):
        for item in dicts:
            a = round(np.random.random(), 2)

            a_item = item(a=a)

            assert repr(a_item) == ' a: ' + str(a)



# -*- coding: utf-8 -*-
"""
Created on  by Ismail Baris
"""
from __future__ import division

import numpy as np

from rspy.ancillary.trigonometry import *

n = 100
DEG = np.random.uniform(0, 360, n)
RAD = np.random.uniform(0, 2 * np.pi, n)


class TestConversion:
    def test_2r_2d(self):
        rads = d2r(DEG)
        np_rads = np.deg2rad(DEG)

        for i, item in enumerate(rads):
            assert np.allclose(item, np_rads[i])

        degs = r2d(RAD)
        np_degs = np.rad2deg(RAD)

        for i, item in enumerate(degs):
            assert np.allclose(item, np_degs[i])

    def test_other(self):
        secs = sec(RAD)

        for i, item in enumerate(secs):
            assert np.allclose(item, 1 / np.cos(RAD[i]))

        cots = cot(RAD)

        for i, item in enumerate(cots):
            assert np.allclose(item, 1 / np.tan(RAD[i]))



# -*- coding: utf-8 -*-
"""
Created on 07.04.19 by ibaris
"""
from __future__ import division

import numpy as np
import pytest

from rspy.constants import optical_range
from rspy.sensor import Sensor
from rspy.sensor.auxiliary import SensorResult

n = 50

dicts = [SensorResult]


class TestAllDicts:
    def test_dicts(self):
        for item in dicts:
            a = np.random.random()

            a_item = item(a=a)

            assert a_item.a == a

    def test_dicts_err(self):
        for item in dicts:
            a = np.random.random()

            a_item = item(a=a)

            with pytest.raises(AttributeError):
                assert a_item.b

    def test_dicts_dir(self):
        for item in dicts:
            a = np.random.random()

            a_item = item(a=a)

            assert dir(a_item) == ['a']

    def test_repr(self):
        for item in dicts:
            a = round(np.random.random(), 2)

            a_item = item(a=a)

            assert repr(a_item) == ' a: ' + str(a)


class TestSatellites:
    def test_L(self):
        test_values = optical_range

        S = Sensor(optical_range, 10, 10, 180, unit='nm', output='THz')
        L8 = S.L8(test_values)

        assert L8.B2 == 482.0
        assert L8.B3 == 561.5
        assert L8.B4 == 654.5
        assert L8.B5 == 865.0
        assert L8.B6 == 1608.5
        assert L8.B7 == 2200.5

    def test_aster(self):
        test_values = optical_range

        S = Sensor(optical_range, 10, 10, 180, unit='nm', output='THz')
        A = S.Aster(test_values)

        assert A.B1 == 560.0
        assert A.B2 == 660.0
        assert A.B3 == 810.0
        assert A.B4 == 1650.0
        assert A.B5 == 2165.0
        assert A.B6 == 2205.0
        assert A.B7 == 2260.0
        assert A.B8 == 2330.0
        assert A.B9 == 2395.0

    def test_excption(self):
        test_values = np.zeros(5)

        S = Sensor(optical_range, 10, 10, 180, unit='nm', output='THz')

        with pytest.raises(AssertionError):
            A = S.Aster(test_values)

            assert A == 0

        with pytest.raises(AssertionError):
            A = S.L8(test_values)

            assert A == 0

    def test_L_ndvi(self):
        test_values = optical_range

        S = Sensor(optical_range, 10, 10, 180, unit='nm', output='THz')
        L8 = S.L8(test_values)
        ndvi = S.ndvi(L8)
        r = L8.B4
        n = L8.B5

        ndvi2 = S.ndvi(red=r, nir=n)

        assert np.allclose(ndvi, 0.13853241197762423)
        assert np.allclose(ndvi, ndvi2)

        with pytest.raises(ValueError):
            assert S.ndvi()

        with pytest.raises(TypeError):
            assert S.ndvi(sensor='kaka')

        with pytest.raises(AssertionError):
            sensor = {'pipi': 'kaka'}
            assert S.ndvi(sensor=sensor)

        with pytest.raises(AssertionError):
            sensor = {'name': 'mumu'}
            assert S.ndvi(sensor=sensor)

    def test_A_ndvi(self):
        test_values = optical_range

        S = Sensor(optical_range, 10, 10, 180, unit='nm', output='THz')
        A = S.Aster(test_values)
        ndvi = S.ndvi(A)
        r = A.B2
        n = A.B3

        ndvi2 = S.ndvi(red=r, nir=n)

        assert np.allclose(ndvi, 0.10204081632653061)
        assert np.allclose(ndvi, ndvi2)

        with pytest.raises(ValueError):
            assert S.ndvi()

        with pytest.raises(TypeError):
            assert S.ndvi(sensor='kaka')

        with pytest.raises(AssertionError):
            sensor = {'pipi': 'kaka'}
            assert S.ndvi(sensor=sensor)

        with pytest.raises(AssertionError):
            sensor = {'name': 'mumu'}
            assert S.ndvi(sensor=sensor)

    def test_L_sr(self):
        test_values = optical_range

        S = Sensor(optical_range, 10, 10, 180, unit='nm', output='THz')
        L8 = S.L8(test_values)
        ndvi = S.sr(L8)
        r = L8.B4
        n = L8.B5

        ndvi2 = S.sr(red=r, nir=n)

        assert np.allclose(ndvi, 1.3216195569136746)
        assert np.allclose(ndvi, ndvi2)

        with pytest.raises(ValueError):
            assert S.ndvi()

        with pytest.raises(TypeError):
            assert S.ndvi(sensor='kaka')

        with pytest.raises(AssertionError):
            sensor = {'pipi': 'kaka'}
            assert S.ndvi(sensor=sensor)

        with pytest.raises(AssertionError):
            sensor = {'name': 'mumu'}
            assert S.ndvi(sensor=sensor)

    def test_A_sr(self):
        test_values = optical_range

        S = Sensor(optical_range, 10, 10, 180, unit='nm', output='THz')
        A = S.Aster(test_values)
        ndvi = S.sr(A)
        r = A.B2
        n = A.B3

        ndvi2 = S.sr(red=r, nir=n)

        assert np.allclose(ndvi, 1.2272727272727273)
        assert np.allclose(ndvi, ndvi2)

        with pytest.raises(ValueError):
            assert S.ndvi()

        with pytest.raises(TypeError):
            assert S.ndvi(sensor='kaka')

        with pytest.raises(AssertionError):
            sensor = {'pipi': 'kaka'}
            assert S.ndvi(sensor=sensor)

        with pytest.raises(AssertionError):
            sensor = {'name': 'mumu'}
            assert S.ndvi(sensor=sensor)



# -*- coding: utf-8 -*-
"""
Created on 07.04.19 by ibaris
"""
from __future__ import division

import numpy as np
import pytest

from rspy.ancillary import *
from rspy.intensity import *

n = 50


class TestBRFBRDF:
    def test_BRF(self):
        for x in range(n):
            BRDF = np.random.uniform(0.00001, 0.1, n)
            I = Intensity(BRDF, value_unit="BRDF")

            assert np.allclose(I.BRF, BRDF * np.pi)
            assert np.allclose(I.I, BRDF)
            assert np.allclose(I.BSC, 0)
            assert np.allclose(I.BSCdB, 0)
            assert I.keys() == ['I', 'BRF', 'BSCdB', 'BSC']

    def test_BRDF(self):
        for x in range(n):
            BRF = np.random.uniform(0.00001, 0.1, n)
            I = Intensity(BRF, value_unit="BRF")

            assert np.allclose(I.BRF, BRF)
            assert np.allclose(I.I, BRF / np.pi)
            assert np.allclose(I.BSC, 0)
            assert np.allclose(I.BSCdB, 0)

    def test_BRDF_ndim3(self):
        for x in range(n):
            BRF = np.random.uniform(0.00001, 0.1, (n, 5, 5))
            I = Intensity(BRF, value_unit="BRF")

            assert np.allclose(I.BRF, BRF)
            assert np.allclose(I.I, BRF / np.pi)
            assert np.allclose(I.BSC, 0)
            assert np.allclose(I.BSCdB, 0)

    def test_exc(self):
        BRF = np.random.uniform(0.00001, 0.1, n)
        with pytest.raises(ValueError):
            assert Intensity(BRF, value_unit="mukmuk")


class TestBSC:
    def test_BSCDEG(self):
        for x in range(n):
            vza = np.random.uniform(10, 50, n)
            BRDF = np.random.uniform(0.00001, 0.1, n)
            I = Intensity(BRDF, value_unit="BRDF", vza=vza, angle_unit='DEG')

            assert np.allclose(I.BRF, BRDF * np.pi)
            assert np.allclose(I.I, BRDF)
            assert np.allclose(I.BSC, BRDF * np.cos(d2r(vza)) * 4 * np.pi)
            assert np.allclose(I.BSCdB, dB(BRDF * np.cos(d2r(vza)) * 4 * np.pi))

    def test_BSCDEG_ndim3(self):
        for x in range(n):
            vza = np.random.uniform(10, 50, n)
            BRDF = np.random.uniform(0.00001, 0.1, (n, 5, 5))
            I = Intensity(BRDF, value_unit="BRDF", vza=vza, angle_unit='DEG')

            BSC_ref = np.empty_like(BRDF)

            for i in range(BSC_ref.shape[0]):
                BSC_ref[i] = BRDF[i] * np.cos(d2r(vza[i])) * 4 * np.pi

            assert np.allclose(I.BRF, BRDF * np.pi)
            assert np.allclose(I.I, BRDF)
            assert np.allclose(I.BSC, BSC_ref)
            assert np.allclose(I.BSCdB, dB(BSC_ref))

    def test_BSCDEG_ndim3_exec(self):
        for x in range(n):
            vza = np.random.uniform(10, 50, n)
            BRDF = np.random.uniform(0.00001, 0.1, (n+1, 5, 5))

            with pytest.raises(AssertionError):
                I = Intensity(BRDF, value_unit="BRDF", vza=vza, angle_unit='DEG')


    def test_BSCRAD(self):
        for x in range(n):
            vza = np.random.uniform(10, 50, n)
            vza = d2r(vza)

            BRDF = np.random.uniform(0.00001, 0.1, n)
            I = Intensity(BRDF, value_unit="BRDF", vza=vza, angle_unit='RAD')

            assert np.allclose(I.BRF, BRDF * np.pi)
            assert np.allclose(I.I, BRDF)
            assert np.allclose(I.BSC, BRDF * np.cos(vza) * 4 * np.pi)
            assert np.allclose(I.BSCdB, dB(BRDF * np.cos(vza) * 4 * np.pi))


class TestBSCINPUT:
    def test_BSCNoVZA(self):
        for x in range(n):
            BSC = np.random.uniform(0.00001, 0.1, n)
            I = Intensity(BSC, value_unit="BSC")

            assert np.allclose(I.BSC, BSC)
            assert np.allclose(I.I, np.zeros_like(I.BSC))
            assert np.allclose(I.BRF, np.zeros_like(I.BSC))

    def test_BSCDEG(self):
        for x in range(n):
            vza = np.random.uniform(10, 50, n)
            BSC = np.random.uniform(0.00001, 0.1, n)
            I = Intensity(BSC, value_unit="BSC", vza=vza, angle_unit='DEG')

            assert np.allclose(I.BSC, BSC)
            assert np.allclose(I.I, BSC / (np.cos(d2r(vza)) * (4 * np.pi)))

    def test_BSCRAD(self):
        for x in range(n):
            vza = np.random.uniform(10, 50, n)
            vza = d2r(vza)
            BSC = np.random.uniform(0.00001, 0.1, n)
            I = Intensity(BSC, value_unit="BSC", vza=vza, angle_unit='RAD')

            assert np.allclose(I.BSC, BSC)
            assert np.allclose(I.BSCdB, dB(BSC))
            assert np.allclose(I.I, BSC / (np.cos(vza) * (4 * np.pi)))
            assert np.allclose(I.BRF, I.I * np.pi)


class TestBSCdBINPUT:
    def test_BSCDEGNoVZA(self):
        for x in range(n):
            BSC = np.random.uniform(0.00001, 0.1, n)
            I = Intensity(dB(BSC), value_unit="BSCdB")

            assert np.allclose(I.BSC, BSC)
            assert np.allclose(I.BSCdB, dB(BSC))
            assert np.allclose(I.I, np.zeros_like(I.BSC))
            assert np.allclose(I.BRF, np.zeros_like(I.BSC))

    def test_BSCDEG(self):
        for x in range(n):
            vza = np.random.uniform(10, 50, n)
            BSC = np.random.uniform(0.00001, 0.1, n)
            I = Intensity(dB(BSC), value_unit="BSCdB", vza=vza, angle_unit='DEG')

            assert np.allclose(I.BSC, linear(dB(BSC)))
            assert np.allclose(I.I, I.BSC / (np.cos(d2r(vza)) * (4 * np.pi)))
            assert np.allclose(I.BRF, I.I * np.pi)

    def test_BSCRAD(self):
        for x in range(n):
            vza = np.random.uniform(10, 50, n)
            vza = d2r(vza)
            BSC = np.random.uniform(0.00001, 0.1, n)
            I = Intensity(BSC, value_unit="BSC", vza=vza, angle_unit='RAD')

            assert np.allclose(I.BSC, BSC)
            assert np.allclose(I.I, BSC / (np.cos(vza) * (4 * np.pi)))
            assert np.allclose(I.BRF, I.I * np.pi)


class TestBRFBRDF:
    def test_BRF(self):
        for x in range(n):
            BRDF = np.random.uniform(0.00001, 0.1, 1)
            I = Intensity(BRDF, value_unit="BRDF")

            vals = [BRDF, BRDF * np.pi, np.zeros(1), np.zeros(1)]

            assert 'I' in I.keys()
            assert 'BRF' in I.keys()
            assert 'BSCdB' in I.keys()
            assert 'BSC' in dir(I)
            assert 'I' in dir(I)
            assert 'BRF' in dir(I)
            assert 'BSCdB' in dir(I)
            assert 'BSC' in dir(I)
            assert len(dir(I)) == 4
            assert len(I.keys()) == 4

            # assert np.allclose(I.values(), vals)
            assert I[0] == BRDF
            assert I[1] == BRDF * np.pi
            assert I[2] == np.zeros(1)
            assert I[3] == np.zeros(1)

            assert I['I'] == BRDF
            assert I['BRF'] == BRDF * np.pi
            assert I['BSC'] == np.zeros(1)
            assert I['BSCdB'] == np.zeros(1)



# -*- coding: utf-8 -*-
"""
Created on 31.03.2019 by Ismail Baris
"""
from __future__ import division

import sys

import numpy as np

from rspy.ancillary import same_len, r2d, d2r, sec, align_all, asarrays
from rspy.auxiliary import valid_angle_deg, valid_angle_rad, check_angle_unit, valid_dtype
from rspy.auxiliary.types import __UNIT_RAD__, __UNIT_DEG__, __DTYPES__
from rspy.constants import pi

# python 3.6 comparability
if sys.version_info < (3, 0):
    srange = xrange
else:
    srange = range

__all__ = ['Angles']


class Angles(object):
    """ Angle Management System

    Angle is a class that helps you unify the different angles of the scanning geometry.

    Attributes
    ----------
    iza, vza, raa, iaa, vaa, alpha, beta: array_like
        SIncidence (iza) and scattering (vza) zenith angle, relative azimuth (raa) angle, incidence and viewing
        azimuth angle (ira, vra) in [RAD].
    Angles.izaDeg, vzaDeg, raaDeg, iaaDeg, vaaDeg, alphaDeg, betaDeg: array_like
        SIncidence (iza) and scattering (vza) zenith angle, relative azimuth (raa) angle, incidence and viewing
        azimuth angle (ira, vra) in [DEG].
    Angles.phi : array_like
        Relative azimuth angle in a range between 0 and 2pi.
    Angles.B, BDeg : array_like
        The result of (1/cos(vza)+1/cos(iza)).
    Angles.mui, muv : array_like
        Cosine of iza and vza in [RAD].
    Angles.geometries : tuple
        If raa is defined it shows a tuple with (iza, vza, raa, alpha, beta) in [RAD]. If iaa and vaa is defined
        the tuple will be (iza, vza, iaa, vaa, alpha, beta) in [RAD]
    Angles.geometriesDeg : tuple
        If raa is defined it shows a tuple with (iza, vza, raa, alpha, beta) in [DEG]. If iaa and vaa is defined
        the tuple will be (iza, vza, iaa, vaa, alpha, beta) in [DEG]
    Angles.nbar : float
        The sun or incidence zenith angle at which the isotropic term is set
        to if normalize is True. You can change this attribute within the class.
    Angles.normlaize : bool
        Set to 'True' to make kernels 0 at nadir view illumination. Since all implemented kernels are normalized
        the default value is False.
    Angles.dtype : numpy.dtype
        Desired data type of all values. This attribute is changeable.

    Methods
    -------
    align_with : Expand all input values to the same length depend on an external array.

    """

    def __init__(self, iza, vza, raa=None, iaa=None, vaa=None, alpha=0.0, beta=0.0, normalize=False, nbar=0.0,
                 angle_unit='DEG', align=True, dtype=np.double):
        """
        Angle is a class that helps you unify the different angles of the scanning geometry.

        Parameters
        ----------
        iza, vza, raa, iaa, vaa : int, float or array_like
            Incidence (iza) and scattering (vza) zenith angle, relative azimuth (raa) angle, incidence and viewing
            azimuth angle (ira, vra). If raa is defined, ira and vra are not mandatory.
        alpha, beta: int, float or array_like
            The Euler angles of the particle orientation (degrees).
        normalize : boolean, optional
            Set to 'True' to make kernels 0 at nadir view illumination. Since all implemented kernels are normalized
            the default value is False.
        nbar : float, optional
            The sun or incidence zenith angle at which the isotropic term is set
            to if normalize is True. The default value is 0.0.
        angle_unit : {'DEG', 'RAD', 'deg', 'rad'}, optional
            * 'DEG': All input angles (iza, vza, raa) are in [DEG] (default).
            * 'RAD': All input angles (iza, vza, raa) are in [RAD].
        align : boolean, optional
             Expand all input values to the same length (default).
        dtype : data-type
            Desired data type of all values. Default is np.double.

        Note
        ----
        Hot spot direction is vza == iza and raa = 0.0

        """

        # Prepare Input Data -------------------------------------------------------------------------
        if raa is None and (iaa is None or vaa is None):
            raise ValueError("If raa is not defined iaa AND vaa must be defined.")

        if raa is not None and iaa is not None and vaa is not None:
            raise AssertionError("The relative, incidence and viewing azimuth angle is defined. "
                                 "Either raa or iaa AND vaa must be defined.")

        if check_angle_unit(angle_unit):
            pass
        else:
            raise ValueError("angle_unit must be {0} or {1}, "
                             "but the actual angle_unit is: {2}".format(str(__UNIT_RAD__),
                                                                        str(__UNIT_DEG__),
                                                                        str(angle_unit)))

        if valid_dtype(dtype):
            pass
        else:
            raise ValueError("Parameter dtype must be {0}. "
                             "The actual dtype is {1}".format(str(__DTYPES__), str(dtype)))

        # Assign relative azimuth angle flag
        if raa is None:
            raa_flag = False
            iaa = iaa
            vaa = vaa
            raa = iaa - vaa

        else:
            raa_flag = True
            raa = raa
            iaa = np.zeros_like(raa)
            vaa = np.zeros_like(raa)

        if align:
            iza, vza, raa, iaa, vaa, alpha, beta = align_all((iza, vza, raa, iaa, vaa, alpha, beta))

        else:
            iza, vza, raa, iaa, vaa, alpha, beta = asarrays((iza, vza, raa, iaa, vaa, alpha, beta), dtype=dtype)

        temporal_array = np.asarray([iza, vza, raa, iaa, vaa, alpha, beta])

        # Check if all data has the same length
        if not same_len(temporal_array):
            raise AssertionError("Input dimensions must agree (try option `align`). The actual dimensions are "
                                 "iza: {0}, vza: {1}, raa: {2}, iaa: {3}, vaa: {4}, "
                                 "alpha: {5} and beta: {6}".format(str(len(iza)), str(len(vza)), str(len(raa)),
                                                                   str(len(iaa)), str(len(vaa)), str(len(alpha)),
                                                                   str(len(beta))))

        # Convert Angles depending on Angle Unit -----------------------------------------------------
        nbar = np.asanyarray(nbar).flatten()

        self.__normalize = normalize

        if valid_angle_deg(angle_unit):
            self.array = d2r(temporal_array)
            self.__nbar = d2r(nbar)

            self.arrayDeg = temporal_array
            self.__nbarDeg = nbar

        elif valid_angle_rad(angle_unit):
            self.array = temporal_array
            self.__nbar = nbar

            self.arrayDeg = r2d(temporal_array)
            self.__nbarDeg = r2d(nbar)

        # Normalize Angles depending on Parameter normalize ------------------------------------------
        self.array = self.__normalize_angles(self.array, self.__nbar[0])
        self.arrayDeg = self.__normalize_angles(self.arrayDeg, self.__nbarDeg[0])

        # Check if there are negative angle values

        negative_angle_list = [0, 1, 5, 6]

        for number in negative_angle_list:
            mask = np.where(self.array[number] < 0)[0]

            if number == 0:
                iza_mask = mask
            elif number == 1:
                vza_mask = mask

            self.array[number][mask] = np.abs(self.array[number][mask])
            self.arrayDeg[number][mask] = np.abs(self.arrayDeg[number][mask])

        self.mask = (iza_mask >= 0) & (vza_mask >= 0)

        for item in self.array[2:-2]:
            item[self.mask] += pi

        for item in self.arrayDeg[2:-2]:
            item[self.mask] += 180.

        # Set Attributes -----------------------------------------------------------------------------
        self.__norm = None
        self.__dtype = dtype

        self.angle_unit = angle_unit
        self.align = align

    # --------------------------------------------------------------------------------------------------------
    # Magic Methods
    # --------------------------------------------------------------------------------------------------------
    def __getitem__(self, key):
        return getattr(self, key)

    # --------------------------------------------------------------------------------------------------------
    # Property Access
    # --------------------------------------------------------------------------------------------------------
    # Access to Array Specific Attributes --------------------------------------------------------------------
    @property
    def len(self):
        """
        Length of array

        Returns
        -------
        len : int
        """
        return self.array.shape[1]

    @property
    def shape(self):
        """
        Shape of array

        Returns
        -------
        shape : tuple
        """
        return self.array.shape

    # Access to Angles -------------------------------------------------------------------------------------------------
    @property
    def iza(self):
        """
        Access the zenith angle of incidence [RAD].

        Returns
        -------
        iza : array_like
        """
        return self.array[0]

    @property
    def izaDeg(self):
        """
        Access the zenith angle of incidence [DEG].

        Returns
        -------
        iza : array_like
        """
        return self.arrayDeg[0]

    @property
    def vza(self):
        """
        Access the zenith angle in viewing direction [RAD].

        Returns
        -------
        vza : array_like
        """
        return self.array[1]

    @property
    def vzaDeg(self):
        """
        Access the zenith angle in viewing direction [DEG].

        Returns
        -------
        vzaDeg : array_like
        """
        return self.arrayDeg[1]

    @property
    def raa(self):
        """
        Access the relative azimuth angle [RAD].

        Note
        ----
        If iaa and raa is defined the relative azimuth angle is calculated like iaa - vaa.

        Returns
        -------
        raa : array_like
        """
        return self.array[2]

    @property
    def raaDeg(self):
        """
        Access the relative azimuth angle [DEG].

        Note
        ----
        If iaa and raa is defined the relative azimuth angle is calculated like iaa - vaa.

        Returns
        -------
        raaDeg : array_like
        """
        return self.arrayDeg[2]

    @property
    def iaa(self):
        """
        Access the azimuth angle of incidence [RAD].

        Note
        ----
        If only raa is defined the iaa and vaa are set to zero.

        Returns
        -------
        iaa : array_like
        """
        return self.array[3]

    @property
    def iaaDeg(self):
        """
        Access the azimuth angle of incidence [DEG].

        Note
        ----
        If only raa is defined the iaa and vaa are set to zero.

        Returns
        -------
        iaaDeg : array_like
        """
        return self.arrayDeg[3]

    @property
    def vaa(self):
        """
        Access the azimuth angle in viewing direction [RAD].

        Note
        ----
        If only raa is defined the iaa and vaa are set to zero.

        Returns
        -------
        vaa : array_like
        """
        return self.array[4]

    @property
    def vaaDeg(self):
        """
        Access the azimuth angle of incidence [DEG].

        Note
        ----
        If only raa is defined the iaa and vaa are set to zero.

        Returns
        -------
        vaaDeg : array_like
        """
        return self.arrayDeg[4]

    @property
    def alpha(self):
        """
        Access the Euler angle alpha of the particle orientation [RAD].

        Returns
        -------
        alpha : array_like
        """
        return self.array[5]

    @property
    def alphaDeg(self):
        """
        Access the Euler angle alpha of the particle orientation [DEG].

        Returns
        -------
        alphaDeg : array_like
        """
        return self.arrayDeg[5]

    @property
    def beta(self):
        """
        Access the Euler angle beta of the particle orientation [RAD].

        Returns
        -------
        beta : array_like
        """
        return self.array[6]

    @property
    def betaDeg(self):
        """
        Access the Euler angle beta of the particle orientation [DEG].

        Returns
        -------
        betaDeg : array_like
        """
        return self.arrayDeg[6]

    @property
    def B(self):
        """
        Access to the sum of the secants of the incidence and scattering angle [RAD].
        The calculation is like: sec(iza) + sec(vza)

        Returns
        -------
        B : array_like
        """
        B = sec(self.iza) + sec(self.vza)
        return B

    @property
    def BDeg(self):
        """
        Access to the sum of the secants of the incidence and scattering angle [DEG].
        The calculation is like: sec(izaDeg) + sec(vzaDeg)

        Returns
        -------
        BDeg : array_like
        """
        B = sec(self.izaDeg) + sec(self.vzaDeg)
        return B

    @property
    def mui(self):
        """
        Access the cosine zenith angle of incidence [RAD].

        Returns
        -------
        mui : array_like
        """
        mui = np.cos(self.iza)
        return mui

    @property
    def muv(self):
        """
        Access the cosine zenith angle in viewing direction [RAD].

        Returns
        -------
        mui : array_like
        """
        muv = np.cos(self.vza)
        return muv

    @property
    def phi(self):
        """
        Relative azimuth angle normalized in a range of 2*pi

        Returns
        -------
        phi : array_like
        """
        phi = np.abs((self.raa % (2. * pi)))
        return phi

    @property
    def geometries(self):
        """
        Access the geometries as tuple objects [RAD].

        Returns
        -------
        geometries : tuple
        """
        geometries = [tuple(self.array[:, i]) for i in srange(self.shape[1])]
        return tuple(geometries)

    @property
    def geometriesDeg(self):
        """
        Access the geometries as tuple objects [DEG].

        Returns
        -------
        geometriesDeg : tuple
        """
        geometriesDeg = [tuple(self.arrayDeg[:, i]) for i in srange(self.shape[1])]
        return tuple(geometriesDeg)

    # ------------------------------------------------------------------------------------------------------------------
    # Property with Setter
    # ------------------------------------------------------------------------------------------------------------------
    # Conversion Routines ----------------------------------------------------------------------------------------------
    @property
    def dtype(self):
        """
        Access the dtype.

        Returns
        -------
        dtype : numpy.dtype
        """
        return self.__dtype

    @dtype.setter
    def dtype(self, value):
        """
        Define a new data type.

        Parameters
        ----------
        value : numpy.dtype

        Returns
        -------
        None
        """
        if valid_dtype(value):
            pass
        else:
            raise ValueError("Parameter dtype must be {0}. "
                             "The actual dtype is {1}".format(str(__DTYPES__), str(value)))

        self.__dtype = value

        self.array = self.__change_dtype(self.array, self.__nbar[0], self.__dtype)
        self.arrayDeg = self.__change_dtype(self.arrayDeg, self.__nbarDeg[0], self.__dtype)

    @property
    def nbar(self):
        """
        Access to the normalization factor nbar [RAD].

        Returns
        -------
        nbar : float
        """
        return self.__nbar

    @nbar.setter
    def nbar(self, value):
        """
        Define a new parameter for nbar.
        After the definition is done, the angles are normalized again as soon as the parameter 'normalize' is True.

        Parameters
        ----------
        value : float
            New nbar value in [RAD].

        Returns
        -------
        None
        """
        self.__nbar = np.asanyarray(value).flatten()
        self.__nbarDeg = r2d(self.__nbar)

        if self.normalize:
            self.array[0][-1] = self.__nbar
            self.arrayDeg[0][-1] = self.__nbarDeg

    @property
    def nbarDeg(self):
        """
        Access to the normalization factor nbar [DEG].

        Returns
        -------
        nbar : float
        """
        return self.__nbarDeg

    @nbarDeg.setter
    def nbarDeg(self, value):
        """
        Define a new parameter for nbar.
        After the definition is done, the angles are normalized again as soon as the parameter 'normalize' is True.

        Parameters
        ----------
        value : float
            New nbar value in [RAD].

        Returns
        -------
        None
        """
        self.__nbarDeg = np.asanyarray(value).flatten()
        self.__nbar = d2r(self.__nbarDeg)

        if self.normalize:
            self.array[0][-1] = self.__nbar
            self.arrayDeg[0][-1] = self.__nbarDeg

    @property
    def normalize(self):
        """
        Access to normalization.

        Returns
        -------
        normalize : bool
        """
        return self.__normalize

    @normalize.setter
    def normalize(self, value):
        """
        Define a new parameter for normalize.
        If value is True, the angles are normalized again as soon as the parameter 'normalize' is False.
        Otherwise (value is False), the angles are de-normalized again as soon as the parameter 'normalize' is True.

        Parameters
        ----------
        value : bool

        Returns
        -------
        None
        """

        if isinstance(value, bool):
            pass
        else:
            raise TypeError("Only bool type can be assigned.")

        if value:
            if self.normalize:
                pass
            else:
                self.__normalize = value
                self.array = self.__normalize_angles(self.array, self.__nbar[0])
                self.arrayDeg = self.__normalize_angles(self.arrayDeg, self.__nbarDeg[0])

        else:
            if not self.normalize:
                pass
            else:
                self.__normalize = value
                self.array = self.__denormalize_angles(self.array)
                self.arrayDeg = self.__denormalize_angles(self.arrayDeg)

    # --------------------------------------------------------------------------------------------------------
    # Callable Methods
    # --------------------------------------------------------------------------------------------------------
    def align_with(self, value):
        """
        Align all angles with another array.

        Parameters
        ----------
        value : array_like

        Returns
        -------
        value : array_like
            Align value.

        Note
        ----
        If len(value) > Angles.shape[1] then the angles inside Angles class will be aligned and it has no effect on
        value. If len(value) < Angles.shape[1] the output of value will be have the same len as Angles and it has no
        effect on the angles within the Angles class.
        """
        # RAD Angles
        data = [item for item in self.array]

        if isinstance(value, (tuple, list)):
            data = tuple(value) + tuple(data, )
        else:
            data = (value,) + tuple(data, )

        data = align_all(data)

        # DEG Angles
        dataDeg = [item for item in self.arrayDeg]

        if isinstance(value, (tuple, list)):
            dataDeg = tuple(value) + tuple(dataDeg, )
        else:
            dataDeg = (value,) + tuple(dataDeg, )

        dataDeg = align_all(dataDeg)

        self.array = np.asarray(data[-7:])
        self.arrayDeg = np.asarray(dataDeg[-7:])

        return data[0:-7]

    # ------------------------------------------------------------------------------------------------------------------
    # Private Methods
    # ------------------------------------------------------------------------------------------------------------------
    def __normalize_angles(self, array, nbar):
        if self.normalize:
            self.__norm = np.array([[nbar], [0], [0], [0], [0], [0], [0]])
            return np.append(array, self.__norm, axis=1)

        return array

    def __denormalize_angles(self, array):
        if not self.normalize:
            return np.delete(array, np.s_[-1:], axis=1)

        return array

    def __change_dtype(self, array, nbar, dtype):
        array = array.astype(dtype)
        return self.__normalize_angles(array, nbar)



# # -*- coding: utf-8 -*-
# """
# Created on 06.04.19 by ibaris
# """
# from __future__ import division
#
# import numpy as np
#
# from rspy.units.ancillary import Units
# from rspy.units.auxiliary import Zero
#
#
# class Quantity(np.ndarray):
#     """ Physical Quantities
#
#     This module defines the `Quantity` objects. A Quantity object represents a number with some an associated unit.
#
#     Attributes
#     ----------
#     Quantity.value : np.ndarray
#         The numerical value of this quantity in the units given by unit.
#     Quantity.unit : sympy.physics.units.quantities.Quantity
#         An object that represents the unit associated with the input value.
#     Quantity.dtype : type
#         The data type of the value
#     Quantity.copy: bool
#         The entered copy bool value.
#     Quantity.order : str
#         Order of the array.
#     Quantity.subok : bool
#         The entered subok value.
#     Quantity.ndmin : int
#         Minimum number of dimensions
#     Quantity.name : str
#         Name of the Quantity
#     Quantity.constant : bool
#         Information about if the Quantity is an constant or not.
#     Quantity.unitstr : str
#         Parameter unit as str.
#     Quantity.unit_mathstr : str
#         Parameter unit as math text.
#     Quantity.label : str
#         Parameter name and unit as math text.
#     Quantity.expr : np.ndarray
#         The whole expression (value * unit) as sympy.core.mul.Mul.
#     Quantity.tolist : list
#         Value and unit as a list.
#
#     Methods
#     -------
#     decompose()
#         Return value as np.ndarray and unit as sympy.physics.units.quantities.Quantity object.
#     decompose_expr(expr)
#         Extract value and unit from a sympy.core.mul.Mul object.
#     set_name(name)
#         Set a name for the current Quantity.
#     convert_to(unit, inplace=True)
#         Convert unit to another units.
#     set_constant(bool)
#         Set a `Quantity` object as constant. In this case, every operation will drop the name of the Quantity.
#     Raises
#     ------
#     UnitError
#     DimensionError
#
#     See Also
#     --------
#     respry.units.Units
#
#     """
#
#     def __new__(cls, value, unit=None, dtype=None, copy=True, order=None,
#                 subok=False, ndmin=0, name=None, constant=True, verbose=False):
#         """
#         The Quantity object is meant to represent a value that has some unit associated with the number.
#
#         Parameters
#         ----------
#         value : float, int, numpy.ndarray, sympy.core.all_classes
#             The numerical value of this quantity in the units given by unit.
#
#         unit : sympy.physics.units.quantities.Quantity, str
#             An object that represents the unit associated with the input value.
#             Must be an `sympy.physics.units.quantities.Quantity` object or a string parsable by
#             the :mod:`~respy.units` package.
#
#         dtype : numpy.dtype, type, int, float, double, optional
#             The dtype of the resulting Numpy array or scalar that will
#             hold the value.  If not provided, it is determined from the input,
#             except that any input that cannot represent float (integer and bool)
#             is converted to float.
#
#         copy : bool, optional
#             If `True` (default), then the value is copied.  Otherwise, a copy will
#             only be made if ``__array__`` returns a copy, if value is a nested
#             sequence, or if a copy is needed to satisfy an explicitly given
#             ``dtype``.  (The `False` option is intended mostly for internal use,
#             to speed up initialization where a copy is known to have been made.
#             Use with care.)
#
#         order : {'C', 'F', 'A'}, optional
#             Specify the order of the array.  As in `~numpy.array`.  This parameter
#             is ignored if the input is a `Quantity` and ``copy=False``.
#
#         subok : bool, optional
#             If `False` (default), the returned array will be forced to be a
#             `Quantity`.
#
#         ndmin : int, optional
#             Specifies the minimum number of dimensions that the resulting array
#             should have.  Ones will be pre-pended to the shape as needed to meet
#             this requirement.  This parameter is ignored if the input is a
#             `Quantity` and ``copy=False``.
#         name : str
#             A name for the created Quantity.
#         constant : bool
#             If True and the constant has a name the name will be replaced after a operation.
#
#         """
#
#         x = np.array(value, dtype=dtype, copy=copy, order=order, subok=subok, ndmin=ndmin)
#         x = np.atleast_1d(x)
#
#         if x.dtype == int:
#             dtype = np.double
#             x = x.astype(dtype)
#         else:
#             pass
#
#         obj = x.view(type=cls)
#
#         if unit is None:
#             obj.unit = Zero
#         else:
#             obj.unit = Units.get_unit(unit)
#
#         obj.__dimension = obj.unit.dimension
#
#         obj.value = x
#
#         if name is None:
#             obj.name = b''
#         else:
#             obj.name = name
#
#         obj.constant = constant
#         obj.dtype = dtype
#         obj.copy = copy
#         obj.order = order
#         obj.subok = subok
#         obj.ndmin = ndmin
#         obj.__quantity__ = True
#         obj.verbose = verbose
#
#         return obj
#
#     # --------------------------------------------------------------------------------------------------------
#     # Magic Methods
#     # --------------------------------------------------------------------------------------------------------
#     def __repr__(self):
#         prefix = '<{0} '.format(self.__class__.__name__)
#         sep = ', '
#         arrstr = np.array2string(self,
#                                  separator=sep,
#                                  prefix=prefix)
#
#         if self.name is None or self.name is b'':
#             return '{0}{1} [{2}]>'.format(prefix, arrstr, Units.unit2str(self.unit))
#
#         else:
#             return '{0}{1} {2} in [{3}]>'.format(prefix, arrstr, self.name, Units.unit2str(self.unit))
#
#     def __array_finalize__(self, obj):
#         if obj is None:
#             return
#         else:
#             self.value = getattr(obj, 'value', None)
#             self.unit = getattr(obj, 'unit', None)
#             self.name = getattr(obj, 'name', None)
#             self.constant = getattr(obj, 'constant', None)
#             self.copy = getattr(obj, 'copy', None)
#             self.order = getattr(obj, 'order', None)
#             self.subok = getattr(obj, 'subok', None)
#             self.ndmin = getattr(obj, 'ndmin', None)
#             self.dimension = getattr(obj, 'dimension', None)
#             self.__quantity__ = getattr(obj, '__quantity__', None)
#             self.verbose = getattr(obj, 'verbose', None)
#
#     def __getitem__(self, item):
#         value = super(Quantity, self).__getitem__(item)
#
#         return self.__new_instance(value, self.unit)
#
#     # --------------------------------------------------------------------------------------------------------
#     # Properties
#     # --------------------------------------------------------------------------------------------------------
#
#     # --------------------------------------------------------------------------------------------------------
#     # Private Methods
#     # --------------------------------------------------------------------------------------------------------
#     def __setattr(self, value, unit, dtype, copy, order, subok, ndmin, name, constant, verbose):
#
#         self.value = value
#         self.dtype = dtype
#         self.copy = copy
#         self.order = order
#         self.subok = subok
#         self.ndmin = ndmin
#         self.name = name
#         self.constant = constant
#         self.verbose = verbose
#         self.unit = unit
#         self.dimension = unit.dimension
#
#     def __new_instance(self, value, unit=None, dtype=None, copy=True, order=None, subok=None, ndmin=None, name=None,
#                        constant=None, verbose=None):
#         quantity_subclass = self.__class__
#
#         unit = self.unit if unit is None else unit
#         dtype = self.dtype if dtype is None else dtype
#         copy = self.copy if copy is None else copy
#         order = self.order if order is None else order
#         subok = self.subok if subok is None else subok
#         ndmin = self.ndmin if ndmin is None else ndmin
#
#         if name is None:
#             name = None if self.constant else self.name
#
#         constant = self.constant if constant is None else constant
#         verbose = self.verbose if verbose is None else verbose
#
#         array = np.array(value, dtype=dtype, copy=copy, order=order, ndmin=ndmin)
#         array = np.atleast_1d(array)
#
#         view = array.view(quantity_subclass)
#         view.__setattr(array, unit, dtype, copy, order, subok, ndmin, name, constant, verbose)
#         view.__array_finalize__(view)
#
#         if subok:
#             return view.value
#
#         return view
#
# # a = np.zeros((3, 2))
# #
# # q = Quantity([1, 2, 3, 4, 5], 'm', name='pipi', subok=True)
# # b = q[0:3]
# #
# # q = Quantity(a.transpose(), 'm', name='pipi', subok=True)



# -*- coding: utf-8 -*-
"""
Created on  by Ismail Baris
"""
from __future__ import division

from numpy import sqrt
from scipy.stats import pearsonr

__all__ = ['pearson', 'rmsd', 'urmsd', 'rmse', 'urmse', 'bias']


def pearson(x, y, square=False):
    """
    Compute the pearson correlation based on scipy.stats.pearsonr

    Parameters
    ----------
    x : numpy.ndarray, int or float
        First sequence.
    y : numpy.ndarray, int or float
        Second sequence.
    square : bool
        If True the output the pearson correlation will be squared, which returns
        the `R-Squared` value.

    Returns
    -------
    out : float

    """
    correlation = pearsonr(x, y)[0]

    if square:
        return correlation ** 2

    return correlation


def bias(x, y):
    return ((x - y) / len(x)).mean()


def rmse(x, y):
    return sqrt(((x - y) ** 2).mean())


def rmsd(x, y):
    return sqrt(((x - y) ** 2 / len(x)).mean())


def urmsd(x, y):
    rsmd = rmsd(x, y)
    b = bias(x, y)

    return sqrt(rsmd ** 2 - b ** 2)


def urmse(x, y):
    rsmd = rmse(x, y)
    b = bias(x, y)

    return sqrt(rsmd ** 2 - b ** 2)



import rspy.bin
import rspy.units
from rspy.ancillary import *
from rspy.angles import Angles
from rspy.auxiliary import *
from rspy.constants import *
from rspy.phase import Phase
from rspy.sensor import Sensor
from rspy.specials.stats import *
from rspy.units import Units
from rspy.waves import Waves
from rspy.intensity import *



# -*- coding: utf-8 -*-
"""
Created on 01.04.2019 by Ismail Baris
"""
from __future__ import division

import os
from distutils import dir_util

import numpy as np
import pytest

from rspy import Phase


@pytest.fixture
def datadir(tmpdir, request):
    """
    Fixture responsible for locating the test data directory and copying it
    into a temporary directory.
    Taken from  http://www.camillescott.org/2016/07/15/travis-pytest-scipyconf/
    """
    filename = request.module.__file__
    test_dir = os.path.dirname(filename)
    data_dir = os.path.join(test_dir, 'data')
    dir_util.copy_tree(data_dir, str(tmpdir))

    def getter(filename, as_str=True):
        filepath = tmpdir.join(filename)
        if as_str:
            return str(filepath)
        return filepath

    return getter


n = 100

iza, vza, iaa, vaa, nbar = np.random.uniform(1, np.pi / 2, n), np.random.uniform(1, np.pi / 2, n), \
                           np.random.uniform(1, 2 * np.pi, n), np.random.uniform(1, 2 * np.pi, n), \
                           np.random.uniform(1, np.pi / 4, 1)

izaDeg, vzaDeg, iaaDeg, vaaDeg, nbarDeg = np.rad2deg(iza), np.rad2deg(vza), np.rad2deg(
    iaa), np.rad2deg(vaa), np.rad2deg(nbar)

raa = iaa - vaa
raaDeg = iaaDeg - vaaDeg

degitems = [izaDeg, vzaDeg, iaaDeg, vaaDeg, raaDeg]
raditems = [iza, vza, iaa, vaa, raa]

degitems_key = ["izaDeg", "vzaDeg", "iaaDeg", "vaaDeg", "raaDeg", "alphaDeg", "betaDeg"]
raditems_key = ["iza", "vza", "iaa", "vaa", "raa", "alpha", "beta"]

property_list = ["len", "shape", "B", "BDeg", "mui", "muv", "phi", "nbar", "nbarDeg", "normalize"]

# "geometries", "geometriesDeg", "array", "arraDeg"

property_list_values = [len(iza), (7, len(iza)), (1 / np.cos(iza) + 1 / np.cos(vza)),
                        (1 / np.cos(izaDeg) + 1 / np.cos(vzaDeg)),
                        np.cos(iza), np.cos(vza), np.abs((raa % (2. * np.pi))), nbar, nbarDeg, False]

array = np.array([iza, vza, raa, iaa, vaa])
arrayDeg = np.array([izaDeg, vzaDeg, raaDeg, iaaDeg, vaaDeg])


class TestProperties:
    def test_properties_and_getitem_conversion_DEG(self):
        align = True
        angle_unit = 'DEG'

        ang = Phase(iza=izaDeg, vza=vzaDeg, iaa=iaaDeg, vaa=vaaDeg,
                    normalize=False, nbar=nbarDeg, angle_unit=angle_unit)

        for i, item in enumerate(degitems):
            assert np.allclose(ang[degitems_key[i]], item)

        for i, item in enumerate(raditems):
            assert np.allclose(ang[raditems_key[i]], item)

        for i, item in enumerate(property_list_values):
            try:
                assert np.allclose(ang[property_list[i]], item)
            except (ValueError, TypeError):
                assert ang[property_list[i]] == item

        assert np.allclose(ang.array[0:-2], array)
        assert np.allclose(ang.arrayDeg[0:-2], arrayDeg)
        assert ang.align == align
        assert ang.angle_unit == angle_unit

        raditems_geom = [iza, vza, raa, iaa, vaa]
        degitems_geom = [izaDeg, vzaDeg, raaDeg, iaaDeg, vaaDeg]

        for i in range(len(iza)):
            radgeom = ang.geometries[i]
            degdgeom = ang.geometriesDeg[i]

            ind = radgeom[0:-2]
            for j, _ in enumerate(ind):
                assert np.allclose(radgeom[j], raditems_geom[j][i])
                assert np.allclose(degdgeom[j], degitems_geom[j][i])

    def test_properties_and_getitem_conversion_RAD(self):
        align = True
        angle_unit = 'RAD'

        ang = Phase(iza=iza, vza=vza, iaa=iaa, vaa=vaa,
                    normalize=False, nbar=nbar, angle_unit=angle_unit)

        for i, item in enumerate(degitems):
            assert np.allclose(ang[degitems_key[i]], item)

        for i, item in enumerate(raditems):
            assert np.allclose(ang[raditems_key[i]], item)

        for i, item in enumerate(property_list_values):
            try:
                assert np.allclose(ang[property_list[i]], item)
            except (ValueError, TypeError):
                assert ang[property_list[i]] == item

        assert np.allclose(ang.array[0:-2], array)
        assert np.allclose(ang.arrayDeg[0:-2], arrayDeg)
        assert ang.align == align
        assert ang.angle_unit == angle_unit

        raditems_geom = [iza, vza, raa, iaa, vaa]
        degitems_geom = [izaDeg, vzaDeg, raaDeg, iaaDeg, vaaDeg]

        for i in range(len(iza)):
            radgeom = ang.geometries[i]
            degdgeom = ang.geometriesDeg[i]

            ind = radgeom[0:-2]
            for j, _ in enumerate(ind):
                assert np.allclose(radgeom[j], raditems_geom[j][i])
                assert np.allclose(degdgeom[j], degitems_geom[j][i])


class TestPhaang:
    def test_phaang(self, datadir):
        fname = datadir("angles.out")

        iza, vza, raa = np.loadtxt(fname, unpack=True)

        fname = datadir("results_dans.out")

        phaang, sinphaang, cosphaang, O, D, piza, pvza = np.loadtxt(fname, unpack=True)

        angle_unit = 'RAD'

        ang = Phase(iza=iza, vza=vza, raa=raa, normalize=False, nbar=nbar, angle_unit=angle_unit)

        assert np.allclose(phaang, ang.phaang)
        assert np.allclose(sinphaang, np.sin(ang.phaang))
        assert np.allclose(cosphaang, np.cos(ang.phaang))

        assert np.allclose(O, ang.O)
        assert np.allclose(D, ang.D)

        phaang_omari = ang.compute_phase_angle(iza, vza, raa, 'Omari')
        li_phaang = ang.compute_Z(method='Li')
        assert np.allclose(phaang_omari, np.arccos(np.cos(iza) * np.cos(vza)))
        assert np.all(li_phaang <= 1)
        assert np.allclose(ang.m, 2)
        assert np.allclose(ang.br, 1)
        assert np.allclose(ang.hb, 2)


class TestException:
    def test_exc(self):
        angle_unit = 'RAD'

        ang = Phase(iza=iza, vza=vza, iaa=iaa, vaa=vaa,
                    normalize=False, nbar=nbar, angle_unit=angle_unit)

        with pytest.raises(ValueError):
            ang.compute_phase_angle(iza, vza, raa, 'pipipopo')

        with pytest.raises(ValueError):
            ang = Phase(iza=iza, vza=vza, iaa=iaa, vaa=vaa,
                        normalize=False, nbar=nbar, angle_unit=angle_unit, method='pupu')

        with pytest.raises(ValueError):
            ang = Phase(iza=iza, vza=vza, iaa=iaa, vaa=vaa,
                        normalize=False, nbar=nbar, angle_unit=angle_unit, orientation='pupu')

        with pytest.raises(NotImplementedError):
            ang = Phase(iza=iza, vza=vza, iaa=iaa, vaa=vaa,
                        normalize=False, nbar=nbar, angle_unit=angle_unit, orientation='vertical')

            ang.compute_Z()

        with pytest.raises(NotImplementedError):
            ang = Phase(iza=iza, vza=vza, iaa=iaa, vaa=vaa,
                        normalize=False, nbar=nbar, angle_unit=angle_unit, orientation='horizontal')

            ang.compute_Z()



# -*- coding: utf-8 -*-
"""
Created on 31.12.2019 by Ismail Baris
"""
from __future__ import division

from numpy import asarray, pad, zeros_like
from numpy import max as npmax

__all__ = ['align_all', 'max_length', 'asarrays', 'get_dtypes', 'same_len',
           'same_shape', 'zeros_likes']


def align_all(data, constant_values='default', dtype=None):
    """
    Align the lengths of arrays.

    Parameters
    ----------
    data : tuple
        A tuple with (mixed) array_like, int, float.
    constant_values : int, float or 'default', optional
        The value at which the smaller values are expand. If 'default' (default) the last value will be choosed.
    dtype : data-type, optional
        Data type of output.

    Returns
    -------
    out : tuple
        Aligned tuple with array_like.
    """
    data = asarrays(data)
    max_len = max_length(data)

    if constant_values == 'default':
        if dtype is None:
            dtypes = get_dtypes(data)

            arrays = list()
            for i, item in enumerate(data):
                padded_array = pad(item, (0, max_len - len(item)), 'constant', constant_values=item[-1])
                arrays.append(asarray(padded_array, dtype=dtypes[i]))

            return arrays

        return asarray(
            [pad(item, (0, max_len - len(item)), 'constant', constant_values=item[-1]) for item in data],
            dtype=dtype)
    else:
        if dtype is None:
            dtypes = get_dtypes(data)

            arrays = list()
            for i, item in enumerate(data):
                padded_array = pad(item, (0, max_len - len(item)), 'constant', constant_values=constant_values)
                arrays.append(asarray(padded_array, dtype=dtypes[i]))

            return arrays

        return asarray(
            [pad(item, (0, max_len - len(item)), 'constant', constant_values=constant_values) for item in data],
            dtype=dtype)


def max_length(data):
    """
    Find the maximum length of the longest object in a tuple.

    Parameters
    ----------
    data : tuple
        A tuple with (mixed) array_like, int, float.

    Returns
    -------
    out : int
    """
    return npmax([len(item) for item in data])


def asarrays(data, dtype=None):
    """
    A wrapper of numpys asarrays for multiple data in a tuple.

    Parameters
    ----------
    data : tuple
        A tuple with (mixed) array_like, int, float.
    dtype : data-type, optional
        Data type of output.

    Returns
    -------
    out : tuple
        A tuple with array_like.
    """
    if dtype is None:
        return [asarray(item).flatten() for item in data]

    return [asarray(item).flatten().astype(dtype) for item in data]


def get_dtypes(data):
    """
    Get dtypes of a tuple with arrays.

    Parameters
    ----------
    data : tuple
        A tuple with (mixed) array_like, int, float.

    Returns
    -------
    out : list
    """
    data = asarrays(data)

    dtypes = list()
    for item in data:
        dtypes.append(item.dtype)

    return dtypes


def same_len(data):
    """
    Determine if the items in a tuple has the same length.

    Parameters
    ----------
    data : tuple
        A tuple with (mixed) array_like, int, float.

    Returns
    -------
    out : bool
    """
    try:
        return all(len(item) == len(data[0]) for item in data)
    except TypeError:
        return False


def same_shape(data):
    """
    Determine if the arrays in a tuple has the same shape.

    Parameters
    ----------
    data : tuple
        A tuple with (mixed) array_like, int, float.

    Returns
    -------
    out : bool
    """
    try:
        return all(item.shape == data[0].shape for item in data)
    except (TypeError, AttributeError):
        return False


def zeros_likes(data, rep=1, dtype=None):
    """
    Get multiple zeros like data.

    Parameters
    ----------
    data : numpy.ndarray
        The shape and data-type of `data` define these same attributes of
        the returned array.
    rep : int
        Determines how many zeros_likes it returns.
    dtype : data-type, optional
        Overrides the data type of the result.

    Returns
    -------
    out : list

    """
    dtype = data.dtype if dtype is None else dtype

    return [zeros_like(data, dtype=dtype) for i in range(rep)]



# -*- coding: utf-8 -*-
"""
Created on  by Ismail Baris
"""
from __future__ import division

import numpy as np
import pytest

from rspy.auxiliary import *
from rspy.auxiliary.types import __DTYPES__, __UNIT_DEG__, __UNIT_RAD__


class TestDtypes:
    def test_white_list(self):
        __white_list__ = __DTYPES__
        __bs_list__ = ['np.bool', np.byte, np.bool, np.ubyte, 'np.ulonglong']

        for item in __white_list__:
            assert valid_dtype(item)

        for item in __bs_list__:
            assert not valid_dtype(item)


class TestAnglesUnits:
    def test_valid_angles_deg(self):
        __white_list__ = __UNIT_DEG__
        __bs_list__ = __UNIT_RAD__

        for item in __white_list__:
            assert valid_angle_deg(item)

        for item in __bs_list__:
            assert not valid_angle_deg(item)

    def test_valid_angles_rad(self):
        __white_list__ = __UNIT_RAD__
        __bs_list__ = __UNIT_DEG__

        for item in __white_list__:
            assert valid_angle_rad(item)

        for item in __bs_list__:
            assert not valid_angle_rad(item)

    def test_check_units(self):
        __white_list__ = __UNIT_DEG__

        for item in __white_list__:
            assert check_angle_unit(item)

        __white_list__ = __UNIT_RAD__

        for item in __white_list__:
            assert check_angle_unit(item)

        __bs_list__ = ['pipidegipipi', 'mumuradimumu', 'radis', 'RadFS', 'DEGGS']

        for item in __bs_list__:
            with pytest.raises(ValueError):
                check_angle_unit(item)



# -*- coding: utf-8 -*-
"""
Created on 03.04.19 by ibaris
"""
from __future__ import division

import numpy as np
from numpy import add, subtract, multiply, true_divide, power
from rspy.auxiliary import Operator

n = 100

operators = (add, subtract, multiply, true_divide, power)

sym_keys = ('+', '-', '*', '/', '**')
class_keys = ('__add__', '__sub__', '__mul__', '__truediv__', '__pow__')
numpy_keys = ('add', 'subtract', 'multiply', 'true_divide', 'power')
KEYS = [sym_keys, class_keys, numpy_keys]


class TestOperator:
    def test_asignment(self):
        for i in range(n):
            select_list = np.random.randint(0, 2)
            select_operator = np.random.randint(0, 4)
            key = KEYS[select_list][select_operator]
            val = operators[select_operator]

            assert Operator[key] == val
            assert Operator.operators[key] == val

    def test_dir(self):
        assert dir(Operator) == ['*',
                                 '**',
                                 '+',
                                 '-',
                                 '/',
                                 '__add__',
                                 '__mul__',
                                 '__pow__',
                                 '__sub__',
                                 '__truediv__',
                                 'add',
                                 'multiply',
                                 'operators',
                                 'power',
                                 'subtract',
                                 'true_divide']



# -*- coding: utf-8 -*-
"""
Created on 07.04.19 by ibaris
"""
from __future__ import division

__all__ = ['SensorResult']


class SensorResult(dict):
    """ Represents the reflectance result.

    Attributes
    ----------
    OpticalResult.name  str
        Name of a sensor or the result.
    OpticalResult.BX : float
        Specific band values for B2 - B7 if OpticalResult.name is 'LANDSAT 8' and B1 - B9 if OpticalResult.name
        is 'ASTER'.

    Notes
    -----
    There may be additional attributes not listed above depending of the
    specific solver. Since this class is essentially a subclass of dict
    with attribute accessors, one can see which attributes are available
    using the `keys()` method.
    """

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __repr__(self):
        if self.keys():
            m = max(map(len, list(self.keys()))) + 1
            return '\n'.join([k.rjust(m) + ': ' + repr(v)
                              for k, v in sorted(self.items())])

        return self.__class__.__name__ + "()"

    def __dir__(self):
        return list(self.keys())
